{$S-,R-,I-,V-,B-,D-,F-}
{$I TADEFINE.INC}
{*********************************************************}
{*                   MEMMAN.PAS 7.00                     *}
{*   Copyright (c) 1988, 1993 by TurboPower Software.    *}
{*                 All rights reserved.                  *}
{*********************************************************}

unit MemMan;
  {-Simplistic memory manager. Assumes that all allocation and
    deallocation is done at the top of the heap (i.e., that
    fragmentation is not possible). DO NOT USE IN PROGRAMS THAT USE
    NORMAL GETMEM/FREEMEM, NEW/DISPOSE, OR MARK/RELEASE CALLS!}

interface

type
  SegOfs =
    record
      Ofst, Segm : Word;
    end;
var
  memLeft : LongInt;

{----- memory management routines -----}

procedure memGet(var P; Size : LongInt);
  {-Allocate a block of Size bytes off the top of the heap}

procedure memFree(Size : LongInt);
  {-Give back a block of Size bytes to the top of the heap. Block must be
    just below HeapPtr^.}

procedure memMark(var P);
  {-Simulate a Mark}

procedure memRelease(P : Pointer);
  {-Deallocate all memory above P}

{----- pointer manipulation routines -----}

function Normalized(P : Pointer) : Pointer;
  {-Return P as a normalized pointer}
  inline(
    $58/                     {pop ax    ;pop offset into AX}
    $5A/                     {pop dx    ;pop segment into DX}
    $89/$C3/                 {mov bx,ax ;BX = Ofs(P^)}
    $B1/$04/                 {mov cl,4  ;CL = 4}
    $D3/$EB/                 {shr bx,cl ;BX = Ofs(P^) div 16}
    $01/$DA/                 {add dx,bx ;add BX to segment}
    $25/$0F/$00);            {and ax,$F ;mask out unwanted bits in offset}

function Linear(P : Pointer) : LongInt;
  {-Converts a pointer to a linear address to allow differences in addresses
    to be calculated. The pointer must be in the range $0:$0 to $FFFF:$000F.}

function LinearToPointer(L : LongInt) : Pointer;
  {-Return linear address L as a normalized pointer}

function PtrDiff(P1, P2 : Pointer) : LongInt;
  {-Return the number of bytes between P1^ and P2^.}

  {=============================================================}

implementation

{$IFNDEF Heap6}
var
  HeapEnd : Pointer;

  function EndOfHeap : Pointer;
    {-Returns a pointer to the end of the free list}
  var
    FreeSegOfs : SegOfs absolute FreePtr;
  begin
    if FreeSegOfs.Ofst = 0 then
      {the free list is empty, add $1000 to the segment}
      EndOfHeap := Ptr(FreeSegOfs.Segm+$1000, 0)
    else
      EndOfHeap := Ptr(FreeSegOfs.Segm+(FreeSegOfs.Ofst shr 4), 0);
  end;
{$ENDIF}

  function Linear(P : Pointer) : LongInt;
    {-Converts a pointer to a linear address to allow differences in addresses
      to be calculated. The pointer must be in the range $0:$0 to $FFFF:$000F.}
  var
    PSO : SegOfs absolute P;
  begin                      {Linear}
    Linear := (LongInt(PSO.Segm) shl 4)+LongInt(PSO.Ofst);
  end;                       {Linear}

  function LinearToPointer(L : LongInt) : Pointer;
    {-Return linear address L as a normalized pointer}
  begin                      {LinearToPointer}
    LinearToPointer := Ptr(Word(L shr 4), Word(L and $0000F));
  end;                       {LinearToPointer}

  function PtrDiff(P1, P2 : Pointer) : LongInt;
    {-Return the number of bytes between P1^ and P2^.}
  begin                      {PtrDiff}
    PtrDiff := Abs(Linear(P1)-Linear(P2));
  end;                       {PtrDiff}

  procedure CalcMemLeft;
    {-Calculate available memory}
  begin                      {CalcMemLeft}
    memLeft := PtrDiff(HeapEnd, HeapPtr);
  end;                       {CalcMemLeft}

  procedure memMark(var P);
    {-Simulate a Mark}
  var
    PP : Pointer absolute P;
  begin                      {memMark}
    PP := HeapPtr;
  end;                       {memMark}

  procedure memRelease(P : Pointer);
    {-Deallocate all memory above P^}
  begin                      {memRelease}
    HeapPtr  := P;
    {$IFDEF Heap6}
    FreeList := P;
    {$ENDIF}
    CalcMemLeft;
  end;                       {memRelease}

  procedure memGet(var P; Size : LongInt);
    {-Allocate Size bytes off the top of the heap}
  var
    PP : Pointer absolute P;
  begin                      {memGet}
    if Size > memLeft then
      PP := nil
    else begin
      {return HeapPtr}
      PP := HeapPtr;

      {set new HeapPtr}
      HeapPtr := LinearToPointer(Linear(HeapPtr)+Size);
      {$IFDEF Heap6}
      FreeList := HeapPtr;
      {$ENDIF}

      {reset memLeft}
      Dec(memLeft, Size);
    end;
  end;                       {memGet}

  procedure memFree(Size : LongInt);
    {-Give back a block of Size bytes to the top of the heap. Block must be
      just below HeapPtr^.}
  begin                      {memFree}
    memRelease(LinearToPointer(Linear(HeapPtr)-Size));
  end;                       {memFree}

begin
  {$IFNDEF Heap6}
  HeapEnd := EndOfHeap;
  {$ENDIF}
  CalcMemLeft;
end.
