{$S-,R-,I-,V-,B-,F-,A-}

{*********************************************************}
{*                  TDATA70.PAS 7.00                     *}
{*    Copyright (c) 1988,1993 by TurboPower Software.    *}
{*                 All rights reserved.                  *}
{*********************************************************}

unit TData70;
  {-Data structures for TInfo70}

interface

type
  SigType = array[1..4] of Char; {a TPU header signature}
  TPUfileType = (TPU7, Other);
  StringPtr = ^string;

const
  SigForTPU70 : SigType = 'TPUQ'; {signature for 7.0 TPx files}
  fpInclude  = 3;            {flags for filename table entries}
  fpUnit     = 4;
  fpObject   = 5;
  fpResource = 6;
  FileTypeTable : array[fpInclude..fpResource] of string[3] = (
    'INC', 'PAS', 'OBJ', 'RES');

  vData     = $00;           {flags for variables}
  vConst    = $01;
  vStack    = $02;
  vAbsolute = $03;
  vVar      = $04;
  vField    = $08;
  vAlias    = $10;

  rSingle   = $00;           {flags for 8087 reals}
  rExtended = $02;
  rDouble   = $04;
  rComp     = $06;

  dSigned   = $01;           {flags for integer types}
  dUnsigned = $02;
  dWord     = $04;
  dLongInt  = $08;

  iShortInt = dSigned;       {integer types}
  iByte     = dUnsigned;
  iInteger  = dWord+dSigned;
  iWord     = dWord+dUnsigned;
  iLongInt  = dLongInt+dWord+dSigned;

  bBoolean  = $00;           {flags for boolean types}
  bByteBool = $10;
  bWordBool = $14;
  bLongBool = $1C;

  ptExport   = $0001;        {procedure table flags}
  ptExtern   = $0002;

  pNear      = $00;          {flags for procedures/methods}
  pFar       = $01;
  pInline    = $02;
  pInterrupt = $04;
  pExternal  = $08;
  pMethod    = $10;
  pConstruct = $20;
  pDestruct  = $40;
  pAssembler = $80;

  cgNoFrame  = $00;          {code generation flags}
  cgExport   = $01;
  cgWinFrame = $02;
  cgDMT      = $04;
  cgHasFrame = $08;
  cgStackMask= cgHasFrame+cgWinFrame;

  ufNPlus    = $0001;       {unit header flags}
  ufOPlus    = $0002;
  ufGPlus    = $0008;
  ufHuh      = $0010; {??}
  ufDOS      = $0000;
  ufDPMI     = $0100;
  ufWindows  = $0200;
  ufUnknown  = $1000; {??}

  scConst    = 'O';          {symbol codes}
  scType     = 'P';
  scVar      = 'Q';
  scProcFunc = 'R';
  scUnit     = 'S';
  scLabel    = 'U';
  scSysProc  = 'V';
  scSysFunc  = 'W';
  scSysPcFc  = 'X';
  scPort     = 'Y';
  scMem      = 'Z';
  scOpenStr  = '[';
  scObject   = 'J';          {made-up symbol codes}
  scFunction = 'F';
type
  Tpu70header =              {format of the TPU header: 7.0 only}
    record
      TPUsig : SigType;      {"TPUQ" signature}
      NextUnit,              {segment in memory for next unit}
      NextLibrary,           {segment in memory for next library}
      UsesPtr,               {offset to unit name/symbol table}
      ScopePtr,              {offset to hash table}
      ProcPtr,               {offset to procedure table}
      GroupPtr,              {offset to Group table}
      ConGrPtr,              {Const group table pointer}
      DatGrPtr,              {Data group table pointer}
      DynaLinkPtr,           {offset to DLL link names table}
      LinkPtr,               {offset to link names table}
      NamePtr,               {offset to filename table}
      LineXlatePtr,          {offset to line number translation table}
      DebugPtr,              {offset to line number table}
      UnitSize,              {symbol table size}
      BrowseSize,            {browser data size}
      CodeSize,              {total code (bytes)}
      ConstSize,             {initialized data (bytes)}
      FixupSize,             {size of code fixup table}
      ConFixSize,            {size of constant fixup section}
      DataSize,              {uninitialized data (bytes)}
      DScopePtr,             {debug scope pointer}
      UnitFlags,             {1 if unit compiled with $N+, 2 if $O+}
      LastObjectPtr,         {pointer to last object in linked list}
      BrowserXrefs,          {offset in browser data for cross-references}
      BrowserStructs : Word; {offset in browser data for structures}
      Private : array[1..21] of Word;
    end;
  TpuPointer = ^Tpu70header;

  GroupRec =                 {an entry in the code group table}
    record
      grOrigin,              {starting offset--filled in by linker}
      grCodeSize,            {size of the code block}
      grFixupSize,           {size of its entry in the fixup table}
      grDebugPtr : Word;     {offset into debug table, or $FFFF}
    end;

  DataGroupRec =             {entry in data or const group tables}
    record
      cdOrigin,              {starting offset--filled in by linker}
      cdSize : Word;         {size of the block}
      cdFixupSize,           {size of its entry in the fixup table}
      cdDebugPtr : Word;     {offset of object symbol if VMT; else 0}
    end;
  DataGroupPtr = ^DataGroupRec;

  LinkNameRec =              {an entry in the link names table}
    record
      kSegment : Word;       {used only at link time}
      kXrefCnt : Word;       {cross-reference count in browser data}
      kName : string;        {name of unit}
    end;

  {if high bit set, value is Word w/ bytes reversed--swap & mask w/ $7FFF}
  StatTable = array[1..1000] of Byte;

  DebugRec =                 {an entry in the line number table}
    record
      deProcPtr,             {offset of group's proc symbol in symbol table}
      deNamePtr,             {offset of filename entry}
      deFirstLine,           {line number, first line of source for procedure}
      deCodeOrg,             {code offset for first line source}
      deLineOrg,             {line number, first line of exec code in procedure}
      deLineCnt : Word;      {number of elements in deStatTab}
      deStatTab : StatTable; {bytes for this statement}
    end;

  SymHeadRec =               {head of an entry in the symbol table}
    record
      LinkCode : Word;       {link to next symbol}
      SymCode : Char;        {symbol type code}
      SymName : string;      {name of the symbol}
    end;
  SymHeadPtr = ^SymHeadRec;

  TypeRec =
    record
      tsType : Pointer;      {pointer to actual type definition}
    end;

  SymTailRec = TypeRec;      {tail of a symbol table entry}
  SymTailPtr = ^SymTailRec;

  UnitNameRec =              {part of a unit name symbol}
    record
      usSegment : Word;      {relative segment for this unit}
      usVersion : Word;      {a checksum}
      usNext : Word;         {offset of symbol for next unit}
      usPrev : Word;         {offset of symbol for previous unit}
    end;

  NameRec =                 {an entry in the filename table}
    record
      nClass : Byte;         {type of the file--see FileTypeTable}
      nIndex : Word;         {used by linker}
      nFileDate : LongInt;   {time stamp for the file}
      nFileName : string;    {name of the file}
    end;

  FixupRec =                 {an entry in the fixup table}
    record
      fxFlags,               {bit flags indicating fixup type}
      fxIndex,               {meaning depends on fxFlags}
      fxOffset,              {"}
      fxLocation : Word;     {"}
    end;

  ProcTableRec =             {an entry in the procedure table}
    record
      prDunno,               {???}
      prFlags,               {flags}
      prGroup,               {an index into the Group table}
      prOffset : Word;       {constant data at start of the block}
    end;
  ProcTablePtr = ^ProcTableRec;

  VarRec =                   {a variable declaration}
    record
      vFlags : Byte;         {vData, vConst, etc.}
      vAddress : Pointer;    {address of variable}
      vHuh : Word;           {???}
      vType : Pointer;       {pointer to type definition}
    end;
  VarRecPtr = ^VarRec;

  FieldRec =                 {tail of entry for a record field}
    record
      vFlags : Byte;         {flags}
      vOffset,               {relative offset of the field}
      vParent,               {offset of parent type}
      vNext : Word;          {link to next field}
      vType : Pointer;       {type definition}
    end;

  ConstRec =                 {entry for a constant}
    record
      cType : Pointer;       {pointer to type definition}
      cValue : LongInt;      {value--not always a longint!}
    end;

  TypeClass = (tUntyped, tArray, tRecord, tObject, tFile, tText, tProc,
               tSet, tPointer, tString, tFloat, tReal, tInteger, tBoolean,
               tChar, tScalar, tUnknown);

  RawTypePtr = ^RawTypeRec;
  RawTypeRec =               {general form of an entry for a type}
    record
      tClass : TypeClass;    {type of symbol}
      tFlags : Byte;         {special flags}
      tSize  : Word;         {size of data type}
      tNext  : Pointer;
      tMinVal,               {minimum value for data type}
      tMaxVal : LongInt;     {maximum value for data type}
      tBase : Pointer;       {pointer to type definition for base type}
      {the following is for enumerated types only--used to construct sets}
      tClass2,
      tFlags2 : Byte;
      tSize2 : Word;
      tNext2 : Word;
      tBase2 : Pointer;
      tfiller : word;         {unknown, added in TP6 }
      tNextField : Byte;
    end;

  ArrayRec =                 {entry for an array}
    record
      tClass : TypeClass;    {= tArray}
      tFlags : Byte;         {special flags}
      tSize  : Word;         {size of data type}
      tNext  : Pointer;
      tComp  : Pointer;      {type def for component}
      tIndex : Pointer;      {type def for array indices}
    end;

  TextRec =                  {entry for a text file}
    record
      tClass : TypeClass;    {= tText}
      tFlags : Byte;         {special flags}
      tSize  : Word;         {size of data type}
      tNext  : Pointer;
      tComp  : Pointer;      {type def for component}
    end;

  LinkTable =
    record
      HashMask : Word;       {number of links in hash table}
      Links : array[1..100] of Word; {arbitrary}
    end;

  ObjectRec =                {entry for an object}
    record
      tClass : TypeClass;    {= tObject}
      tFlags : Byte;         {special flags}
      tSize : Word;          {size of data type}
      tNameOfs : Word;       {offset of entry for object's name}
      tHuh : Word;
      tScope : Word;         {offset of local scope (link table)}
      tFieldPtr : Word;      {offset of entry for first field/method}
      tParent : Pointer;     {pointer to type def of parent object}
      tMethSize : Word;      {size of VMT table}
      tMethGroup : Word;     {const group index for VMT}
      tMethOfs : Word;       {offset in object for VMT pointer}
      tDMTGroup : Word;      {const group index for DMT}
      tPrevObj : Word;       {offset of previous object in linked list}
      tHuh1 : Word;
      tHuh2 : Word;
      tHuhPtr1 : Pointer;
      tHuhPtr2 : Pointer;
      tHashMask : Word;      {links to fields/methods}
      tLinks : array[1..8] of Word;
    end;

  RecordRec =                {entry for a record}
    record
      tClass : TypeClass;    {= tRecord}
      tFlags : Byte;         {special flags}
      tSize : Word;          {size of data type}
      tParent : Word;        {offset of parent}
      tDunno : Word;         {??}
      tScope : Word;         {offset of local scope (link table)}
      tFieldPtr : Word;      {offset of entry for first field}
    end;

  ParamRec =                 {entry for a parameter}
    record
      rType : Pointer;       {pointer to type def for parameter}
      rFlags : Byte;         {vStack or vStack+vVar}
    end;

  ProcFuncRec =              {entry for a procedure variable}
    record
      tClass : TypeClass;    {= tProc}
      tFlags : Byte;         {special flags}
      tSize : Word;          {size of data type}
      tHuh  : Word;          {??}
      tNext : Word;          {offset of next method }
      tResType : Pointer;    {type def for function result, or nil for proc}
      tParCnt : Word;        {number of parameters}
      tParams : array[1..1000] of ParamRec; {parameters}
    end;

  ProcRec =                  {entry for a procedure, function, method}
    record
      pFlags : Byte;         {pFar, pInline, etc.}
      pCodeGenFlags : Byte;  {code generation flags: cgXXXX}
      pIndex,                {index into proc table or size of code for macro}
      pParent,               {offset of parent}
      pScope,                {offset of local link table}
      pMethOfs : Word;       {offset in VMT table (non 0 = virtual)}
      pType : ProcFuncRec;   {parameters, etc.}
    end;

  SysProcRec =               {entry for a system procedure/function}
    record
      JumpTableIndex : Word;
    end;

  PortRec =                  {entry for Port/PortW}
    record
      poSize : Byte;
    end;

  MemRec =                   {entry for Mem/MemW}
    record
      meType : Pointer;
    end;

  LabelRec =                 {entry for a label}
    record
      aTarget : Word;        {?}
      aJumpList : Word;      {?}
    end;

  SegOfs =                   {structure of a pointer}
    record
      Ofst, Segm : Word;
    end;

  Array4 = array[0..4] of Byte;

const
  ClassTypes : array[TypeClass] of string[11] = (
    'untyped', 'array', 'record', 'object', 'File', 'Text', 'Procedure',
    'Set', 'Pointer', 'String', '8087 real', 'Real', 'scalar',
    'Boolean', 'Char', 'enumerated', 'unknown');

var
  Buffer : Pointer;          {pointer to contents of the TPU file}
  CodeSeg : Word;            {segment for the code}
  RelSeg : Word;             {segment for the relocation table}
  BaseSeg : Word;            {segment for first TPU}
  SystemSeg     : Word;
  Con : Text;
  ConHandle : Word absolute Con;
  ConBuf : Byte;
  ThisFileType : TPUfileType; {the file type}
const
  SymNest : Word = 0;        {keeps track of symbol nesting for reporting}
  PlusMinus : array[Boolean] of Char = ('-', '+');
  RecursionMax = 100;
var
  {deal with recursion}
  RecursionStack : array[1..RecursionMax] of Pointer;
const
  RecursionPtr : Word = 0;

procedure Error(Msg : string);
  {-Display an error message and halt}

procedure PushAddr(P : Pointer);
  {-Push an address onto the recursion stack}

procedure PopAddr;
  {-Pop an address off the recursion stack}

function Recurring(P : Pointer) : Boolean;
  {-Return True if we're recurring}

function SymbolClass(Code : Char) : string;
  {-Return the class name for a symbol code}

function ClassifyFile(Header : TpuPointer) : TPUfileType;
  {-Classify FName as TPU, TPL, or Other}

function Paras(Bytes : Word) : Word;
  {-Convert bytes to paragraphs}

  {==========================================================================}

implementation

  procedure Error(Msg : string);
    {-Display an error message and halt}
  begin
    WriteLn(Con, Msg);
    Halt(1);
  end;

  procedure PushAddr(P : Pointer);
    {-Push an address onto the recursion stack}
  begin
    if RecursionPtr = RecursionMax then
      Error('Too much recursion!');
    Inc(RecursionPtr);
    RecursionStack[RecursionPtr] := P;
  end;

  procedure PopAddr;
    {-Pop an address off the recursion stack}
  begin
    Dec(RecursionPtr);
  end;

  function Recurring(P : Pointer) : Boolean;
    {-Return True if we're recurring}
  var
    I : Word;
  begin
    Recurring := False;
    for I := 1 to RecursionPtr do
      if RecursionStack[I] = P then begin
        Recurring := True;
        Exit;
      end;
  end;

  function SymbolClass(Code : Char) : string; {*OK*}
    {-Return the class name for a symbol code}
  begin
    case Code of
      'O' : SymbolClass := 'constant';
      'P' : SymbolClass := 'type';
      'Q' : SymbolClass := 'variable/typed const';
      'R' : SymbolClass := 'procedure';
      'S' : SymbolClass := 'unit/program name';
      'U' : SymbolClass := 'label';
      'V' : SymbolClass := 'system procedure';
      'W' : SymbolClass := 'system function';
      'X' : SymbolClass := 'system procedure/function';
      'Y' : SymbolClass := 'port/portw';
      'Z' : SymbolClass := 'mem/memw/meml';
      '[' : SymbolClass := 'openstring';
      else SymbolClass := '? unknown';
    end;
  end;

  function ClassifyFile(Header : TpuPointer) : TPUfileType;
    {-Classify FName as TPU7 or Other}
  var
    ST : SigType;
  begin
    {classify the file}
    ST := Header^.TPUsig;
    if ST = SigForTPU70
      Then ClassifyFile := TPU7
    else
      ClassifyFile := other;
  end;

  function Paras(Bytes : Word) : Word;
    {-Convert bytes to paragraphs}
  begin
    Paras := (Bytes+$000F) shr 4;
  end;

begin
  {initialize globals}
  FillChar(RecursionStack, SizeOf(RecursionStack), 0);
end.
