{*********************************************************}
{*                   TINFO70.PAS 7.00                    *}
{*    Copyright (c) 1988,1993 by TurboPower Software.    *}
{*                 All rights reserved.                  *}
{*********************************************************}

{$S-,R-,I-,V-,B-,F-}
{$M 32768,2000,650000}

{$IFDEF DPMI}
  !! Error: cannot be compiled for a DPMI target
{$ENDIF}
{$IFDEF Windows}
  !! Error: cannot be compiled for a Windows target
{$ENDIF}

program TInfo70;
  {-Display information about a Turbo Pascal 7.0 TPU/TPP/TPW/TPL file}

{$I TADEFINE.INC}

uses
  MemMan,
  TData70,
  TMsg70,
  {$IFDEF UseTPRO}
  TpInline,
  TpString,
  TpDos,
  {$ELSE}
  OpInline,
  OpString,
  OpDos,
  {$ENDIF}
  Unasm,
  TPeek70;

  {.F-}
const
  SignOn =
    'Turbo Pascal 7.0 Info. Copyright 1993 by TurboPower Software. Version 7.00.'^M^J;
  MaxSyms       = 2500;
  StdErrHandle  = 2;
  DoSymbolNames : Boolean = False;
  DoProcSizes   : Boolean = False;
  DoVarTypes    : Boolean = False;
  DoDump        : Boolean = False;
  DoInterface   : Boolean = False;
  DoImplement   : Boolean = False;
  NeedOf        : Boolean = False;
  DoingArray    : Boolean = False;
  BooleanString : array[0..1] of string[5] = ('false', 'true');
  CRLF          : string[2] = ^M^J;
var
  InF           : file;
  FName         : string[64];
  CurrentCol    : Byte;
  CurrentSym    : Word;
  SymArray      : array[1..MaxSyms] of SymHeadPtr;
  SymCount      : Word;
{.F+}

  procedure GetTypeName(TP, TypePtr : Pointer; Search, InsertColon : Boolean);
  forward;

  function SymbolIsTypedConst(PHead : SymHeadPtr) : Boolean;
  forward;

  function SymbolIsEnumConst(TP : Pointer; PHead : SymHeadPtr) : Boolean;
  forward;

  function Real2Str(R : Real; Width, Places : Byte) : string;
    {-Convert a real to a string}
  var
    S : string;
  begin
    Str(R:Width:Places, S);
    Real2Str := S;
  end;

  function HexP(P : Pointer) : string;
    {-Return P in $ssss:$oooo format}
  begin
    with SegOfs(P) do
      HexP := '$'+HexW(Segm)+':$'+HexW(Ofst);
  end;

  procedure WriteIt(S : string);
    {-Write S, keeping track of current column}
  var
    NewCol : Word;
  begin
    NewCol := CurrentCol+Length(S);
    if (S = CRLF) then begin
      WriteLn;
      CurrentCol := 1;
    end
    else if Length(S) = 1 then begin
      Write(S);
      Inc(CurrentCol);
    end
    else case NewCol of
      1..79 : begin
                Write(S);
                CurrentCol := NewCol;
              end;
      else    begin
                Write(CRLF, S);
                CurrentCol := Succ(Length(S));
              end;
    end;
  end;

  function ProcIsFunc(PHead : SymHeadPtr) : Boolean;
    {-Return True if the procedure pointed to by PHead is a function}
  var
    PTail : SymTailPtr;
    PHSO : SegOfs absolute PHead;
    PRP : ^ProcRec absolute PTail;
  begin
    PTail := Ptr(PHSO.Segm, PHSO.Ofst+4+Length(PHead^.SymName));
    ProcIsFunc := PRP^.pType.tResType <> nil;
  end;

  procedure DumpSymbol(PHead : SymHeadPtr; var Count : Word);
    {-Dump an individual symbol}
  begin
    with PHead^ do begin
      Inc(Count);
      if Count > 1 then
        if CurrentCol+Length(SymName)+2 <= 80 then
          WriteIt(', ')
        else
          WriteIt(',');
      WriteIt(SymName);
    end;
  end;

  procedure DumpSymbolTable(TP : Pointer; S : string; TypeCode : Char);
    {-Dump a symbol table}
  var
    SHP : SymHeadPtr;
    I, Count : Word;
    TCode : Char;
  begin
    WriteLn(^M^J, S);
    Count := 0;
    for I := 1 to SymCount do begin
      SHP := SymArray[I];

      {make sure typed constants end up with constants rather than variables}
      TCode := TypeCode;
      if SymbolIsTypedConst(SHP) then begin
        if TCode = scConst then
          TCode := scVar
        else
          TCode := #0;
      end
      {suppress display of enumerated constants}
      else if SymbolIsEnumConst(TP, SHP) then
        TCode := #0
      {separate objects from other types}
      else if Char(Byte(SHP^.SymCode) and $7F) = scType then begin
        if TypeIsObject(SHP) then
          if TCode = scObject then
            TCode := scType
          else
            TCode := #0
        else if TCode = scObject then
          TCode := #0;
      end
      {separate procedures and functions}
      else if Char(Byte(SHP^.SymCode) and $7F) = scProcFunc then begin
        if ProcIsFunc(SHP) then
          if TCode = scFunction then
            TCode := scProcFunc
          else
            TCode := #0
        else if TCode = scFunction then
          TCode := #0;
      end;

      if Char(Byte(SHP^.SymCode) and $7F) = TCode then
        {dump this symbol}
        DumpSymbol(SHP, Count);
    end;
    if Count = 0 then
      WriteIt('(none)');
    WriteIt(CRLF);
  end;

  procedure DumpVarTypes(TP : TpuPointer);
    {-Dump the names of all variables and their types and sizes}
  var
    PHead : SymHeadPtr;
    PTail : SymTailPtr;
    PTSO : SegOfs absolute PTail;
    PVR : ^VarRec absolute PTail;
    I, Count, Size : Word;
    S : string[20];
  begin
    Count := 0;
    WriteLn;
    WriteLn('Type        Class  Size Name');
    WriteLn('----------- ----- ----- ---------------');
    for I := 1 to SymCount do begin
      PHead := SymArray[I];
      if Char(Byte(PHead^.SymCode) and $7F) = scVar then begin
        PTail := Pointer(PHead);
        Inc(PTSO.Ofst, 4+Length(PHead^.SymName));
        Inc(Count);

        S := VarType(TP, PVR^.vType, Size);
        Write(Pad(S, 12), Pad(VarClass(PVR^.vFlags), 6), Size:5, ' ',
          PHead^.SymName);
        if PVR^.vFlags = vAbsolute then
          WriteLn(' (', HexPtr(PVR^.vAddress), ')')
        else
          WriteLn;
      end;
    end;
    if Count = 0 then
      WriteLn('No variables or typed constants found.');
  end;

  procedure FindCodeSizes(TP : TpuPointer; ProcIndex : Word;
                           var BlockSize, ConstData, EstSize : Word);
    {-Given an index into the procedure table, return the size of the entire
      code block and the amount of constant data ahead of the code}
  var
    PTP : ProcTablePtr;
    LRP : ^GroupRec;
    I, LI, TempCD, HigherCD, MaxOffset : Word;
  begin
    SegOfs(PTP).Segm := SegOfs(TP).Segm;
    SegOfs(PTP).Ofst := TP^.ProcPtr+ProcIndex;
    if PTP^.prFlags and ptExtern <> 0 then begin
      BlockSize := 0;
      ConstData := 0;
      EstSize := 0;
      Exit;
    end;
    SegOfs(LRP).Segm := Seg(TP^);
    LI := PTP^.prGroup;
    SegOfs(LRP).Ofst := TP^.GroupPtr+LI;
    BlockSize := LRP^.grCodeSize;
    TempCD := PTP^.prOffset;
    ConstData := TempCD;

    {point at first entry in procedure table}
    SegOfs(PTP).Ofst := TP^.ProcPtr;

    {search for first block with the same link index}
    while PTP^.prGroup <> LI do
      Inc(SegOfs(PTP).Ofst, SizeOf(ProcTableRec));

    {search for an entry at a higher offset in the same block}
    MaxOffset := TP^.GroupPtr; {NamePtr}
    HigherCD := $FFFF;
    while (SegOfs(PTP).Ofst < MaxOffset) do begin
      I := PTP^.prOffset;
      if (PTP^.prGroup = LI) then
        if (I > TempCD) and (I < HigherCD) then
          HigherCD := I;
      Inc(SegOfs(PTP).Ofst, SizeOf(ProcTableRec));
    end;

    {if we found something, use the difference in offsets}
    if HigherCD <> $FFFF then
      EstSize := HigherCD-TempCD
    else
      EstSize := BlockSize-TempCD;
  end;

  procedure DumpProcSizes(TP : TpuPointer);
    {-Dump the names of all procedures/functions and their sizes}
  var
    PHead : SymHeadPtr;
    PTail : SymTailPtr;
    PTSO : SegOfs absolute PTail;
    PR : ^ProcRec absolute PTail;
    PO : ^ObjectRec;
    LHead : SymHeadPtr;
    LHSO : SegOfs absolute LHead;
    LTail : SymTailPtr;
    LTSO : SegOfs absolute LTail;
    FRP : ^FieldRec absolute LTail;
    PRP : ^ProcRec absolute LTail;
    I, Count : Word;
    LCName : string[80];

    procedure DumpPrim(Index : Word; Name : string);
    var
      BlockSize, ConstData, ApproxSize : Word;
    begin
      {calculate sizes}
      FindCodeSizes(TP, Index, BlockSize, ConstData, ApproxSize);
      WriteLn(BlockSize:5, ConstData:6, ApproxSize:6, '  ', Name);
      Inc(Count);
    end;

  begin
    Count := 0;
    WriteLn;
    WriteLn('Block Entry Approx Procedure/');
    WriteLn(' Size Point  Size  Method');
    WriteLn('----- ----- -----  ----------');
    for I := 1 to SymCount do begin
      PHead := SymArray[I];
      PTail := Pointer(PHead);
      Inc(PTSO.Ofst, 4+Length(PHead^.SymName));
      LCName := PHead^.SymName;

      {do procedures, but skip inline macros}
      if (Char(Byte(PHead^.SymCode) and $7F) = scProcFunc) and (PR^.pFlags <> pInline) then
        DumpPrim(PR^.pIndex, LCName)
      {do methods in objects}
      else if (Char(Byte(PHead^.SymCode) and $7F) = scType) and TypeIsObject(PHead) then begin
        PO := PTail^.tsType;
        SegOfs(PO).Segm := PTSO.Segm;
        SegOfs(PO).Ofst := SegOfs(PTail^.tsType).Ofst;

        LHSO.Segm := PTSO.Segm;
        LHSO.Ofst := PO^.tFieldPtr;
        while LHSO.Ofst <> 0 do begin
          LTSO.Segm := PTSO.Segm;
          LTSO.Ofst := LHSO.Ofst+4+Length(LHead^.SymName);

          {is it a field?}
          if Char(Byte(LHead^.SymCode) and $7F) = scVar then
            LHSO.Ofst := FRP^.vNext
          {if not, it's a method}
          else begin
            DumpPrim(PRP^.pIndex, LCName+'.'+LHead^.SymName);
            LHSO.Ofst := PRP^.pType.tNext;
          end;
        end;
      end;
    end;
    if Count = 0 then
      WriteLn('No procedures or methods found.');
  end;

  function FindUnitName(P : Pointer) : Word;
    {-Find unit name}
  var
    I, O : Word;
    SHP : SymHeadPtr;
  begin
    FindUnitName := 0;
    for I := 1 to SymCount do begin
      SHP := SymArray[I];
      with SHP^ do
        if Char(Byte(SymCode) and $7F) = scUnit then begin
          O := SegOfs(SHP).Ofst+4+Length(SymName);
          if O = SegOfs(P).Segm then begin
            FindUnitName := I;
            Exit;
          end;
        end;
    end;
  end;

  function FindTypeName(P : Pointer) : Word;
    {-If the name of the specified type is known, return an index into
      SymArray, else 0}
  var
    I : Word;
    PT : ^TypeRec;
    PTSO : SegOfs absolute PT;
    SHP : SymHeadPtr;
  begin
    FindTypeName := 0;
    if SegOfs(P).Segm <> CurrentUnitSymbol then
      Exit;
    for I := 1 to SymCount do begin
      SHP := SymArray[I];
      with SHP^ do
        if Char(Byte(SymCode) and $7F) = scType then begin
          PT := Pointer(SHP);
          Inc(PTSO.Ofst, 4+Length(SymName));
          if PT^.tsType = P then begin
            FindTypeName := I;
            Exit;
          end;
        end;
    end;
  end;

  {$F+}
  function FindTypeNameString(TP : TpuPointer; P : Pointer; Strict : Boolean) : string;
    {-}
  var
    I, Size : Word;
  begin
    I := FindTypeName(P);
    if I = 0 then
      if Strict and (SegOfs(P).Segm = CurrentUnitSymbol) then
        FindTypeNameString := 'unknown'
      else
        FindTypeNameString := VarType(TP, P, Size)
    else
      FindTypeNameString := SymArray[I]^.SymName;
  end;

  function FindProcNameByIndex(TP : TpuPointer; O : Word) : Boolean;
    {-Find a procedure name by its index into the procedure table}
  var
    I : Word;
    PR : ^ProcRec;
    PRSO : SegOfs absolute PR;
    PTail : SymTailPtr;
    PTSO : SegOfs absolute PTail;
    LHead : SymHeadPtr;
    LHSO : SegOfs absolute LHead;
    LTail : SymTailPtr;
    LTSO : SegOfs absolute LTail;
    FRP : ^FieldRec absolute LTail;
    PRP : ^ProcRec absolute LTail;
    PO : ^ObjectRec;
    SHP : SymHeadPtr;
  begin
    FindProcNameByIndex := False;
    for I := 1 to SymCount do begin
      SHP := SymArray[I];
      with SHP^ do
        case Char(Byte(SymCode) and $7F) of
          scProcFunc :
            begin
              PR := Pointer(SHP);
              Inc(PRSO.Ofst, 4+Length(SymName));
              if (PR^.pFlags <> pInline) and (PR^.pIndex = O) then begin
                Write(SymArray[I]^.SymName);
                FindProcNameByIndex := True;
                Exit;
              end;
            end;
          scType:
            if TypeIsObject(SHP) then begin
              PTail := Pointer(SHP);
              Inc(PTSO.Ofst, 4+Length(SHP^.SymName));
              SegOfs(PO).Segm := PTSO.Segm;
              SegOfs(PO).Ofst := SegOfs(PTail^.tsType).Ofst;

              LHSO.Segm := PTSO.Segm;
              LHSO.Ofst := PO^.tFieldPtr;
              while LHSO.Ofst <> 0 do begin
                LTSO.Segm := PTSO.Segm;
                LTSO.Ofst := LHSO.Ofst+4+Length(LHead^.SymName);

                {is it a field?}
                if Char(Byte(LHead^.SymCode) and $7F) = scVar then
                  LHSO.Ofst := FRP^.vNext
                {if not, it's a method}
                else if PRP^.pIndex = O then begin
                  Write(SHP^.SymName, '.', LHead^.SymName);
                  FindProcNameByIndex := True;
                  Exit;
                end
                else
                  LHSO.Ofst := PRP^.pType.tNext;
              end;
            end;
        end;
    end;
  end;
  {$F-}

  function ValToString(var V; MT : TypeClass; ST : Byte; Size : Word;
                       TP, P : Pointer) : string;
    {-Convert value V to a printable string}
  var
    S : string;
    I : Byte;
    PHead : SymHeadPtr;
    PT : ^RawTypeRec absolute P;
  label
    BackUp;
  begin
    S := '';
    case MT of
      tPointer :                        {pointer}
        if Pointer(V) = nil then
          S := 'nil'
        else
          S := HexP(Pointer(V));
      tString :                        {string}
        begin
          S := string(V);
          for I := 1 to Length(S) do
            if S[I] < ' ' then
              S[I] := '.';
          S := ''''+S+'''';
        end;
      tReal :                        {real}
        S := Real2Str(Real(V), 0, 2);
      tInteger :                     {scalar}
        begin
BackUp:   case ST of
            iShortInt : S := Long2Str(ShortInt(V));
            iByte     : S := Long2Str(Byte(V));
            iInteger  : S := Long2Str(Integer(V));
            iWord     : S := Long2Str(Word(V));
            iLongInt  : S := Long2Str(LongInt(V));
            else
              begin
                case Size of
                  1 : ST := iByte;
                  2 : ST := iWord;
                  else ST := iLongInt;
                end;
                goto BackUp;
              end;
          end;
        end;
      tBoolean :                        {boolean}
        S := BooleanString[Byte(V)];
      tScalar :                        {enumerated}
        if SegOfs(P).Segm = CurrentUnitSymbol then begin
          SegOfs(P).Segm := SegOfs(TP).Segm;
          with PT^ do begin
            PHead := @tNextField;
            for I := 0 to tMaxVal do begin
              if I = Byte(V) then
                S := PHead^.SymName;
              Inc(SegOfs(PHead).Ofst, 12+Length(PHead^.SymName));
            end;
          end;
        end
        else
          S := Long2Str(Byte(V));
      tChar :                        {char}
        case Char(V) of
          #0..#31 :
            S := '^'+Char(Byte(V)+64);
          else
            S := ''''+Char(V)+'''';
        end;
    end;
    if Length(S) = 0 then
      ValToString := '??'
    else
      ValToString := S;
  end;

  procedure WriteChar(cValue : Byte);
    {-Write character value cValue}
  begin
    Write(ValToString(cValue, tChar, 0, 0, nil, nil));
  end;

  function CalcConstAddress(TP : TpuPointer; P : Pointer) : Pointer;
    {-Calculate the address in the constant group section of the data for the
      typed constant associated with P}
  var
    I, J, S, O : Word;
    DGP : DataGroupPtr;
    DSO : SegOfs absolute DGP;
  begin
    {calculate segment}
    with TP^ do
      S := SegOfs(TP).Segm+Paras(UnitSize)+Paras(BrowseSize)+Paras(CodeSize);

    {calculate offset}
    DSO.Segm := SegOfs(TP).Segm;
    DSO.Ofst := TP^.ConGrPtr;
    O := 0;
    J := SegOfs(P).Segm div SizeOf(DataGroupRec);
    for I := 1 to J do begin
      Inc(O, DGP^.cdSize);
      Inc(DSO.Ofst, SizeOf(DataGroupRec));
    end;
    Inc(O, SegOfs(P).Ofst);
    CalcConstAddress := Ptr(S, O);
  end;

  function GetParameterName(TP : TpuPointer; Ofs, Param : Word) : string;
    {-Return the name for the Param'th parameter for the procedure whose
      symbol starts at the specified Offset}
  var
    SHP : SymHeadPtr;
    SHPSO : SegOfs absolute SHP;
    I : Word;
  begin
    GetParameterName := '??';
    if Ofs <> 0 then begin
      SHP := Ptr(SegOfs(TP).Segm, Ofs+10);
      for I := 1 to Param-1 do
        Inc(SHPSO.Ofst, 15+Length(SHP^.SymName));
      if Char(Byte(SHP^.SymCode) and $7F) = scVar then
        GetParameterName := SHP^.SymName;
    end;
  end;

  procedure GetTypePrim(TP : TpuPointer; TA : Pointer; var InsertColon : Boolean);
    {-Write out a declaration for TA}
  const
    Recur : Byte = 0;
  var
    MT : TypeClass;
    ST : Byte;
    TASO : SegOfs absolute TA;
    PT : ^RawTypeRec;
    PTSO : SegOfs absolute PT;
    PX : ^TextRec absolute PT;
    PRR : ^RecordRec absolute PT;
    PA : ^ArrayRec absolute PT;
    PFRP : ^ProcFuncRec absolute PT;
    POR : ^ObjectRec absolute PT;
    PHead : SymHeadPtr;
    PFR : ^FieldRec;
    OMP : ^ProcRec absolute PFR;
    Indent, I, O, Size : Word;
    InPrivateSection : Boolean;
    Code             : Char;

    procedure DoQuestionMarks(P : Pointer);
    var
      I : Word;
    begin
      I := FindUnitName(P);
      if I <> 0 then
        Write(SymArray[I]^.SymName, '.');
      Write('??');
    end;

  begin
    MT := FindTypeClass(TP, TA, Size, ST, PT);

    {don't show structure of objects within other type definitions}
    if (Recur = 1) and (MT = tObject) then
      Exit;

    Inc(Recur);
    Indent := (Recur*2)+2;
    if InsertColon and (MT <> tUntyped) then begin
      Write(' : ');
      InsertColon := False;
    end;

    if NeedOf and (MT <> tArray) then begin
      Write('] of ');
      NeedOf := False;
    end;

    with PT^ do
      case MT of
        tUntyped :                     {untyped}
          {do nothing} ;
        tArray :                     {array}
          begin
            if DoingArray then
              Write(', ')
            else
              Write('array[');
            DoingArray := True;
            NeedOf := False;
            GetTypeName(TP, PA^.tIndex, True, False);
            NeedOf := True;
            GetTypeName(TP, PA^.tComp, True, False);
            DoingArray := False;
          end;
        tRecord :                     {record}
          begin
            O := PRR^.tFieldPtr;
            if O = 0 then
              Write('record end')
            else begin
              WriteLn(^M^J, ' ':Indent, 'record');
              repeat
                PHead := Ptr(PTSO.Segm, O);
                PFR := Ptr(PTSO.Segm, O+Length(PHead^.SymName)+4);
                Write(' ':Indent+2, PHead^.SymName);
                Inc(Recur);
                GetTypeName(TP, PFR^.vType, True, True);
                Dec(Recur);
                WriteLn(';');
                O := PFR^.vNext;
              until (O = 0);
              Write(' ':Indent, 'end');
            end;
          end;
        tFile :                     {file}
          with PX^ do begin
            Write('File');
            if PX = nil then
              MT := tUntyped
            else
              MT := FindTypeClass(TP, tComp, Size, ST, PT);
            if MT <> tUntyped then begin
              Write(' of ');
              GetTypeName(TP, tComp, True, False);
            end;
          end;
        tText, tReal :
          Write(ClassTypes[MT]);
        tProc :                   {procedure/function}
          with PFRP^ do begin
            if tResType = nil then
              Write('procedure')
            else
              Write('function');
            if tParCnt > 0 then begin
              Write('(');
              if tParCnt > 1 then
                WriteLn;
              for I := 1 to tParCnt do
                with tParams[I] do begin
                  if tParCnt > 1 then
                    Write(' ':Indent);
                  if rFlags and vVar <> 0 then
                    Write('var ');
                  Write('??');
                  GetTypeName(TP, rType, True, True);
                  if I < tParCnt then
                    WriteLn(';');
                end;
              Write(')');
            end;
            if tResType <> nil then
              GetTypeName(TP, tResType, True, True);
          end;
        tSet :                     {set}
          begin
            Write('set of ');
            GetTypeName(TP, PX^.tComp, True, False);
          end;
        tPointer :                     {pointer}
          if TASO.Segm = SystemSeg then
            if ST = $FF then
              Write('PChar')
            else
              Write('Pointer')
          else begin
            Write('^');
            GetTypeName(TP, PX^.tComp, True, False);
          end;
        tBoolean :                    {boolean}
          case ST of
            bBoolean  : Write('Boolean');
            bByteBool : Write('ByteBool');
            bWordBool : Write('WordBool');
            bLongBool : Write('LongBool');
          end;
        tString :                     {string}
          begin
            Write('string');
            if Size < 256 then
              Write('[', Size-1, ']');
          end;
        tChar :                     {char}
          if TASO.Segm = SystemSeg then
            Write('Char')
          else begin
            WriteChar(tMinVal);
            Write('..');
            WriteChar(tMaxVal);
          end;
        tScalar :                     {enumerated}
          if tBase = TA then begin
            Write('(');
            if tMaxVal > 3 then
              WriteLn;
            PHead := @tNextField;
            for I := 0 to tMaxVal do begin
              if (I > 0) then begin
                Write(',');
                if tMaxVal > 3 then
                  WriteLn
                else
                  Write(' ');
              end;
              if tMaxVal > 3 then
                Write(' ':Indent);
              Write(PHead^.SymName);
              Inc(SegOfs(PHead).Ofst, 12+Length(PHead^.SymName));
            end;
            Write(')');
          end
          else
            Write(ValToString(tMinVal, MT, ST, Size, TP, tBase), '..',
                  ValToString(tMaxVal, MT, ST, Size, TP, tBase));
        tFloat :                     {8087 real}
          case ST of
            rSingle   : Write('Single');
            rExtended : Write('Extended');
            rDouble   : Write('Double');
            rComp     : Write('Comp');
            else        Write('??');
          end;
        tInteger :                     {scalar}
          begin
            if TASO.Segm <> SystemSeg then
              ST := 0;
            case ST of
              iShortInt : Write('ShortInt');
              iByte     : Write('Byte');
              iInteger  : Write('Integer');
              iWord     : Write('Word');
              iLongInt  : Write('LongInt');
              else        Write(tMinVal, '..', tMaxVal);
            end;
          end;
        tObject :
          with POR^ do begin
            Write(^M^J, ' ':Indent, 'object');
            if tParent = nil then
              WriteLn
            else begin
              Write('(');
              I := FindTypeName(tParent);
              if I = 0 then
                DoQuestionMarks(tParent)
              else
                Write(SymArray[I]^.SymName);
              WriteLn(')');
            end;
            O := tFieldPtr;
            InPrivateSection := false;
            while (O <> 0) do begin
              PHead := Ptr(PTSO.Segm, O);
              PFR := Ptr(PTSO.Segm, O+Length(PHead^.SymName)+4);
              Code := PHead^.SymCode;
              if not InPrivateSection and (Byte(Code) and $80 <> 0) then begin
                WriteLn(' ':Indent, 'private:');
                InPrivateSection := True;
              end
              else if InPrivateSection and (Byte(Code) and $80 = 0) then begin
                WriteLn(' ':Indent, 'public:');
                InPrivateSection := False;
              end;
              Code := Char(Byte(Code) and $7F);
              if Code = scVar then begin
                Inc(Recur);
                Write(' ':Indent+2, PHead^.SymName);
                GetTypeName(TP, PFR^.vType, True, True);
                Dec(Recur);
                O := PFR^.vNext
              end
              else with OMP^, pType do begin
                Write(' ':Indent+2);
                if pFlags and pConstruct <> 0 then
                  Write('constructor')
                else if pFlags and pDestruct <> 0 then
                  Write('destructor')
                else if tResType = nil then
                  Write('procedure')
                else
                  Write('function');
                Write(' ', PHead^.SymName);
                if tParCnt > 0 then begin
                  Write('(');
                  if tParCnt > 1 then
                    WriteLn;
                  for I := 1 to tParCnt do
                    with tParams[I] do begin
                      if tParCnt > 1 then
                        Write(' ':Indent+4);
                      if rFlags and vVar <> 0 then
                        Write('var ');
                      Write(GetParameterName(TP, pScope, I));
                      GetTypeName(TP, rType, True, True);
                      if I < tParCnt then
                        WriteLn(';');
                    end;
                  Write(')');
                end;
                if tResType <> nil then
                  GetTypeName(TP, tResType, True, True);
                if pMethOfs <> 0 then begin
                  Write('; virtual');
                  if pCodeGenFlags and cgDMT <> 0 then begin
                    Write(' ');
                    if TP^.UnitFlags and ufWindows = ufWindows then begin
                      if pMethOfs >= $A000 then
                        Write('cm_First + ', CMtoStr(pMethOfs-$A000))
                      else if pMethOfs >= $9000 then
                        Write('nf_First + ', pMethOfs-$9000)
                      else if pMethOfs >= $8000 then
                        Write('id_First + ', IDtoStr(pMethOfs-$8000))
                      else
                        Write('wm_First + ', WMtoStr(pMethOfs));
                    end
                    else
                      Write(pMethOfs);
                  end;
                end;
                if pFlags and pAssembler <> 0 then
                  Write('; assembler');
                if pFlags and pExternal <> 0 then
                  Write('; external');
                if pCodeGenFlags and cgExport <> 0 then
                  Write('; export');
                O := OMP^.pType.tNext;
              end;
              WriteLn(';');
            end;
            Write(' ':Indent, 'end');
          end;
        else
          DoQuestionMarks(TA);
      end;
    Dec(Recur);
  end;

  procedure GetTypeName(TP, TypePtr : Pointer; Search, InsertColon : Boolean);
    {-Write out a declaration for TypePtr}
  var
    I : Word;
  begin
    if Search then
      I := FindTypeName(TypePtr)
    else
      I := 0;

    if I = 0 then
      GetTypePrim(TP, TypePtr, InsertColon)
    else begin
      if InsertColon then
        Write(' : ');
      if NeedOf then begin
        Write('] of ');
        NeedOf := False;
      end;
      Write(SymArray[I]^.SymName);
    end;
  end;

  function AbsoluteString(VRP : VarRecPtr) : string;
    {-Return the name of the variable that VRP is absolute on top of}
  var
    I, O : Word;
    VRP2 : VarRecPtr;
    VOfs : Word absolute VRP2;
    SHP : SymHeadPtr;
  begin
    AbsoluteString := '??';
    with VRP^ do
      if SegOfs(vAddress).Segm = CurrentUnitSymbol then begin
        O := SegOfs(vAddress).Ofst;
        for I := 1 to SymCount do begin
          SHP := SymArray[I];
          with SHP^ do
            if Char(Byte(SymCode) and $7F) = scVar then begin
              VRP2 := Pointer(SHP);
              Inc(VOfs, 4+Length(SymName));
              if VOfs = O then begin
                AbsoluteString := SymName;
                Exit;
              end;
            end;
         end;
      end
      else begin
        I := FindUnitName(vAddress);
        if I <> 0 then
          AbsoluteString := SymArray[I]^.SymName + '.??';
      end;
  end;

  procedure ReconstructSymbol(TP : TpuPointer; PHead : SymHeadPtr; Code : Char);
    {-Reconstruct the declaration for PHead}
  var
    PTail : SymTailPtr;
    PHSO : SegOfs absolute PHead;
    CRP : ^ConstRec absolute PTail;
    VRP : VarRecPtr absolute PTail;
    PRP : ^ProcRec absolute PTail;
    TRP : ^TypeRec absolute PTail;
    PT : ProcTablePtr;
    PTSO : SegOfs absolute PT;
    P : Pointer;
    I, Size : Word;
    MT : TypeClass;
    ST : Byte;

    function DynaLinkName(O : Word) : string;
    var
      LNP : ^LinkNameRec;
    begin
      LNP := Ptr(Seg(TP^), TP^.DynaLinkPtr+O);
      DynaLinkName := LNP^.kName;
    end;

  begin
    PTail := Ptr(PHSO.Segm, PHSO.Ofst+4+Length(PHead^.SymName));
    case Code of
      scConst :                   {constant}
        with CRP^ do begin
          Write(' = ');
          MT := FindTypeClass(TP, cType, Size, ST, P);
          if MT = tScalar then begin
            MT := tInteger; ST := $D;
          end;
          Write(ValToString(cValue, MT, ST, Size, TP, cType));
        end;
      scType :                    {type}
        with TRP^ do begin
          Write(' = ');
          GetTypeName(TP, tsType, False, False);
        end;
      scVar :                     {variable}
        with VRP^ do begin
          GetTypeName(TP, vType, True, True);
          case vFlags of
            vConst :              {typed constant}
              begin
                MT := FindTypeClass(TP, vType, Size, ST, P);
                P := CalcConstAddress(TP, vAddress);
                Write(' = ');
                if (MT = tString) and (Length(string(P^)) > 40) then
                  Write(^M^J, '    ');
                Write(ValToString(P^, MT, ST, Size, TP, vType));
              end;
            vAbsolute :           {absolute var}
              Write(' absolute ', HexP(vAddress));
            else                  {var absolute on var?}
              if vFlags and vAlias <> 0 then
                Write(' absolute ', AbsoluteString(VRP));
          end;
        end;
      scProcFunc :                {procedure}
        with PRP^, pType do begin
          if tParCnt > 0 then begin
            Write('(');
            if tParCnt > 1 then
              WriteLn;
            for I := 1 to tParCnt do
              with tParams[I] do begin
                if tParCnt > 1 then
                  Write(' ':12);
                if rFlags and vVar <> 0 then
                  Write('var ');
                Write(GetParameterName(TP, pScope, I));
                GetTypeName(TP, rType, True, True);
                if I < tParCnt then
                  WriteLn(';');
              end;
            Write(')');
          end;
          if tResType <> nil then
            GetTypeName(TP, tResType, True, True);

          PTSO.Segm := PHSO.Segm;
          PTSO.Ofst := TP^.ProcPtr+pIndex;

          if pFlags and pExternal <> 0 then
            Write('; external')
          else if pFlags = pInline then
            Write('; inline')
          else if pFlags = pInterrupt then
            Write('; interrupt')
          else if pCodeGenFlags and cgExport <> 0 then
            Write('; export')
          else if PT^.prFlags and ptExtern <> 0 then
            Write('; external ''', DynaLinkName(PT^.prGroup), ''' index ',
                  PT^.prOffset);
        end;
    end;
  end;

  function SymbolIsInterfaced(TPU : TpuPointer; P : Pointer) : Boolean;
    {-Return True if P points to an interfaced symbol}
  var
    Pofs : Word absolute P;
  begin
    with TPU^ do
      if DScopePtr = ScopePtr then
        SymbolIsInterfaced := True
      else
        SymbolIsInterfaced := POfs < DScopePtr;
  end;

  function SymbolIsTypedConst(PHead : SymHeadPtr) : Boolean;
    {-Return True if P points to a typed constant}
  var
    PHSO : SegOfs absolute PHead;
    PTail : SymTailPtr absolute PHead;
    VRP : VarRecPtr absolute PTail;
  begin
    if Char(Byte(PHead^.SymCode) and $7F) <> scVar then
      SymbolIsTypedConst := False
    else begin
      Inc(PHSO.Ofst, 4+Length(PHead^.SymName));
      SymbolIsTypedConst := (VRP^.vFlags = vConst);
    end;
  end;

  function SymbolIsEnumConst(TP : Pointer; PHead : SymHeadPtr) : Boolean;
    {-Return True if P points to a enumerated constant}
  var
    PTSO : SegOfs;
    PTail : SymTailPtr absolute PTSO;
    CRP : ^ConstRec absolute PTail;
    Size : Word;
    ST : Byte;
    Par : ^ParamRec;
  begin
    SymbolIsEnumConst := False;
    if Char(Byte(PHead^.SymCode) and $7F) = scConst then begin
      PTSO := SegOfs(PHead);
      Inc(PTSO.Ofst, 4+Length(PHead^.SymName));
      with CRP^ do
        SymbolIsEnumConst :=
          FindTypeClass(TP, cType, Size, ST, Par) = tScalar;
    end;
  end;

  procedure Reconstruct(TP : Pointer; S : string; TypeCode : Char;
                        Interfaced : Boolean);
    {-Reconstruct interfaced declarations of the specified type}
  var
    I : Word absolute CurrentSym;
    Count : Word;
    TCode : Char;
    SHP : SymHeadPtr;
  begin
    Count := 0;
    for I := 1 to SymCount do begin
      SHP := SymArray[I];

      {make sure typed constants end up with constants rather than variables}
      TCode := TypeCode;
      if SymbolIsTypedConst(SHP) then
        if TCode = scConst then
          TCode := scVar
        else
          TCode := #0;

      {suppress display of enumerated constants}
      if SymbolIsEnumConst(TP, SHP) then
        TCode := #0;

      if Char(Byte(SHP^.SymCode) and $7F) = TCode then
        if SymbolIsInterfaced(TP, SHP) = Interfaced then begin
          Inc(Count);
          if Count = 1 then
            WriteLn;
          case TypeCode of
            scProcFunc :
              if ProcIsFunc(SHP) then
                Write('function ')
              else
                Write(S, ' ');
            else
              begin
                if Count = 1 then
                  WriteLn(S);
                Write('  ');
              end;
          end;
          Write(SHP^.SymName);

          {reconstruct this symbol}
          ReconstructSymbol(TP, SHP, TCode);

          WriteLn(';');
        end;
    end;
  end;

  procedure AddSymbol(SHP : SymHeadPtr);
    {-Add a symbol to our linked list}
  begin
    if SymCount = MaxSyms then begin
      WriteLn('Only ', MaxSyms, ' symbol table entries allowed.');
      Halt(1);
    end;
    Inc(SymCount);
    SymArray[SymCount] := SHP;
  end;

  {$F+}
  procedure ReadSymbolTable(TP : TpuPointer; HashOfst : Word);
    {-Read a symbol table}
  var
    P : SymHeadPtr;
    PSO : SegOfs absolute P;
    Entries, Count,
    UnitSeg, I, O : Word;
  begin
    UnitSeg := Seg(TP^);
    P := Pointer(TP);
    O := HashOfst;
    Entries := MemW[UnitSeg:O] shr 1;
    Count := 0;
    SymCount := 0;
    for I := 0 to Entries do begin
      Inc(O, 2);
      PSO.Ofst := O;
      PSO.Ofst := P^.LinkCode;
      while PSO.Ofst <> 0 do begin
        {add this symbol to the list}
        AddSymbol(P);

        {follow the links}
        PSO.Ofst := P^.LinkCode;
      end;
    end;
  end;
  {$F-}

  procedure SortSymbols(Alpha : Boolean);
    {-Sort the symbols}
  var
    InOrder : Boolean;
    I, J, K, Offset : Word;
  begin
    Offset := SymCount;
    while Offset > 1 do begin
      Offset := Offset shr 1;
      repeat
        InOrder := True;
        K := SymCount-Offset;
        for J := 1 to K do begin
          I := J+Offset;

          {compare I and J}
          if Alpha then begin
            if SymArray[I]^.SymName < SymArray[J]^.SymName then begin
              {swap elements I and J}
              ExchangeStructs(SymArray[I], SymArray[J], SizeOf(Pointer));

              {not in order yet}
              InOrder := False;
            end;
          end
          else begin
            if LongInt(SymArray[I]) < LongInt(SymArray[J]) then begin
              {swap elements I and J}
              ExchangeStructs(SymArray[I], SymArray[J], SizeOf(Pointer));

              {not in order yet}
              InOrder := False;
            end;
          end;
        end;
      until InOrder;
    end;
  end;

  procedure DumpFileTable(TP : TpuPointer);
    {-Dump a filename table}
  var
    P : ^NameRec;
    PSO : SegOfs absolute P;
    StopPoint : Word;
    S : string[30];
  begin
    P := Pointer(TP);
    PSO.Ofst := TP^.NamePtr;
    StopPoint := TP^.LineXlatePtr;
    WriteLn(^M^J'Source Files:');
    while (PSO.Ofst < StopPoint) do
      with P^ do begin
        S := nFileName;
        Inc(PSO.Ofst, Length(nFileName)+8);
        if PSO.Ofst < StopPoint then
          S := S+', ';
        WriteIt(S);
      end;
    WriteIt(CRLF);
  end;

  procedure DumpUsesList(TP : TpuPointer);
    {-Dump a USES list}
  var
    I : Word;
    S : string;
    SHP : SymHeadPtr;
    SHPSO : SegOfs absolute SHP;
    UNP : ^UnitNameRec;
    UNPSO : SegOfs absolute UNP;
  begin
    WriteLn(^M^J'Uses:');

    I := 0;
    SHP := Ptr(SegOfs(TP).Segm, TP^.UsesPtr);
    UNPSO.Segm := SHPSO.Segm;

    repeat
      UNPSO.Ofst := SHPSO.Ofst;
      Inc(UNPSO.Ofst, Length(SHP^.SymName)+4);
      if I <> 0 then begin
        S := SHP^.SymName;
        if UNP^.usNext <> 0 then
          S := S+', ';
        WriteIt(S);
      end;
      SHPSO.Ofst := UNP^.usNext;
      Inc(I);
    until SHPSO.Ofst = 0;

    if I = 0 then
      WriteIt('nothing');
    WriteIt(CRLF);
  end;

  procedure DumpTpuHeader(Header : TpuPointer);
    {-Dump a TPU header}
  var
    UN : StringPtr;
  begin
    with Header^ do begin
      UN := Ptr(SegOfs(Header).Segm, UsesPtr+3);
      WriteLn('Unit name:          ', UN^,
        ^M^J'Unit type:          ', UnitType(Header),
        ^M^J'Total code:         ', CodeSize,
        ^M^J'Initialized data:   ', ConstSize,
        ^M^J'Uninitialized data: ', DataSize,
        ^M^J'Symbol table:       ', UnitSize,
        ^M^J'Fixup table:        ', FixupSize,
        ^M^J'Debug table:        ', UnitSize-DebugPtr,
        ^M^J'Numeric processing: $N', PlusMinus[UnitFlags and ufNPlus <> 0],
        ^M^J'Allow overlay:      $O', PlusMinus[UnitFlags and ufOPlus <> 0],
        ^M^J'286 code:           $G', PlusMinus[UnitFlags and ufGPlus <> 0]);
    end;
  end;

  procedure DumpTPU;
    {-Dump a TPU/TPP/TPW/TPL file}
  var
    P : Pointer;
    TPU : TpuPointer absolute P;
    PSO : SegOfs absolute P;
    SHP : SymHeadPtr;
    Done : Boolean;
    Implement : Boolean;
  begin
    {save current segment}
    P := Buffer;
    repeat
      BaseSeg := PSO.Segm;
      Implement := DoImplement;

      {get offset of current unit symbol}
      CurrentUnitSymbol := TPU^.UsesPtr;
      SHP := Ptr(PSO.Segm, CurrentUnitSymbol);
      Inc(CurrentUnitSymbol, Length(SHP^.SymName)+4);
      SystemSeg := FindSystemSeg(TPU);

      {get symbols if necessary}
      SymCount := 0;
      if DoSymbolNames or DoProcSizes or DoVarTypes
      or DoInterface or DoImplement then begin
        {read the symbol table entries}
        ReadSymbolTable(TPU, TPU^.DScopePtr);
        Implement := Implement and (TPU^.DScopePtr <> TPU^.ScopePtr);

        {sort them alphabetically}
        SortSymbols(True);
      end;

      {dump the header}
      DumpTpuHeader(TPU);

      {dump the uses list}
      DumpUsesList(TPU);

      {dump the file table}
      DumpFileTable(TPU);

      {dump symbols if desired}
      if DoSymbolNames then begin
        DumpSymbolTable(P, 'Constants:',  scConst);
        DumpSymbolTable(P, 'Types:',      scType);
        DumpSymbolTable(P, 'Objects:',    scObject);
        DumpSymbolTable(P, 'Variables:',  scVar);
        DumpSymbolTable(P, 'Procedures:', scProcFunc);
        DumpSymbolTable(P, 'Functions:',  scFunction);
      end;

      {dump procedure sizes if desired}
      if DoProcSizes then
        DumpProcSizes(TPU);

      {dump variable types/sizes if desired}
      if DoVarTypes then
        DumpVarTypes(TPU);

      {reconstruct interface and/or implementation section if desired}
      if DoInterface or Implement then
        {sort them in declaration order}
        SortSymbols(False);

      if DoInterface then begin
        WriteLn(^M^J'interface');
        Reconstruct(P, 'const',     scConst,    True);
        Reconstruct(P, 'type',      scType,     True);
        Reconstruct(P, 'var',       scVar,      True);
        Reconstruct(P, 'procedure', scProcFunc, True);
        WriteLn;
      end;

      if Implement then begin
        WriteLn(^M^J'implementation');
        Reconstruct(P, 'const',     scConst,    False);
        Reconstruct(P, 'type',      scType,     False);
        Reconstruct(P, 'var',       scVar,      False);
        Reconstruct(P, 'procedure', scProcFunc, False);
        WriteLn;
      end;

      {point to next TPU, if any}
      with TPU^ do begin
        Inc(PSO.Segm, Paras(UnitSize)+Paras(CodeSize)+Paras(FixupSize)+
                      Paras(ConstSize));
        Done := (TPU^.TPUsig <> SigForTPU70);
      end;

      if not Done then
        WriteLn(^M^J);
    until Done;
  end;

  procedure ReadFile;
    {-Read the entire TPU file into memory}
  var
    P : Pointer;
    Pofs : Word absolute P;
    BytesRead : Word;
  begin
    P := Buffer;
    repeat
      {read a big chunk}
      {$I-}
      BlockRead(InF, P^, $FF00, BytesRead);
      if IoResult <> 0 then
        Error('Error reading from .TPU file');
      {$I+}

      {point to next byte in buffer}
      Inc(Pofs, BytesRead);

      {normalize the pointer}
      P := Normalized(P);
    until (BytesRead = 0);
    Close(InF);
  end;

  procedure Help;
    {-Display instructions}
  begin
    WriteLn(Con, 'Usage: TINFO70 tpxfile[.TPU] [Options] [>Output]'^M^J);
    WriteLn(Con, 'Options:');
    WriteLn(Con, '  /S  Generate a list of symbol names');
    WriteLn(Con, '  /P  Generate a list of procedures/methods and their sizes');
    WriteLn(Con, '  /V  Generate a list of variables and their types and sizes');
    WriteLn(Con, '  /D  Generate a formatted dump of the TPx file');
    WriteLn(Con, '  /R  Reconstruct interface section');
    WriteLn(Con, '  /I  Reconstruct implementation section');
    WriteLn(Con, '  /A  Select all of the above options');
    WriteLn(Con, '  /?  Display this help screen');
    Halt(0);
  end;

  procedure ReadCommandLine;
    {-Get command line parameters}
  var
    I : Byte;
    Arg : string;

    procedure Unrecognized(Msg : string);
      {-Report error for unrecognized command line option}
    begin
      Error('Unrecognized option: '+Msg);
    end;

  begin
    {show help screen if no command line parameters are given}
    if ParamCount = 0 then
      Help;

    I := 1;
    while I <= ParamCount do begin
      Arg := ParamStr(I);
      if (Arg[1] = '/') or (Arg[1] = '-') then begin
        if Length(Arg) <> 2 then
          Unrecognized(Arg);
        case Upcase(Arg[2]) of
          'S' : DoSymbolNames := not DoSymbolNames;
          'P' : DoProcSizes := not DoProcSizes;
          'V' : DoVarTypes := not DoVarTypes;
          'R' : DoInterface := not DoInterface;
          'I' : DoImplement := not DoImplement;
          'D' : DoDump := not DoDump;
          'A' :
            begin
              DoSymbolNames := True;
              DoProcSizes := True;
              DoVarTypes := True;
              DoInterface := True;
              DoImplement := True;
              DoDump := True;
            end;
          '?' : Help;
          else  Unrecognized(Arg);
        end
      end
      else if Arg = '?' then
        Help
      else if Length(FName) = 0 then
        FName := Arg
      else
        Unrecognized(Arg);
      Inc(I);
    end;

    {check for a data file}
    if Length(FName) = 0 then
      Error('No .TPx file specified. Enter "TINFO70" for help.');
  end;

  procedure Initialize;
    {-Initialize globals, get command line parameters, etc.}
  var
    TotalBytes : LongInt;
    SaveName : string;
    HasExt : Boolean;

    function FindFile(Ext : string) : Boolean;
    begin
      FName := DefaultExtension(SaveName, Ext);
      FindFile := ExistOnPath(FName, FName);
    end;

  begin
    CurrentCol := 1;
    FName := '';

    {get command line parameters}
    ReadCommandLine;

    {check for existence of FName}
    FName := StUpcase(CleanPathName(FName));
    SaveName := FName;
    HasExt := JustExtension(FName) <> '';
    FName := DefaultExtension(FName, 'TPU');
    if not ExistOnPath(FName, FName) then
      if HasExt then
        Error(SaveName+' not found')
      else if (not FindFile('TPP')) and (not FindFile('TPW')) and (not FindFile('TPL')) then
        Error(SaveName+' not found');

    {open the input file}
    Assign(InF, FName);
    Reset(InF, 1);

    {allocate one big block of memory}
    TotalBytes := FileSize(InF);
    memGet(Buffer, TotalBytes);
    if Buffer = nil then
      Error('Insufficient memory');

    {read the file into the buffer}
    ReadFile;

    {determine the kind of file}
    ThisFileType := ClassifyFile(Buffer);
    if ThisFileType <> TPU7 then begin
      FName := JustFileName(FName);
      Error(FName+' is not a version 7.0 TPU, TPP, TPW, or TPL file');
    end;
  end;

begin
  FindTypeNameFunc := FindTypeNameString;
  FindProcNameByIndexFunc := FindProcNameByIndex;
  ReadSymbolTableProc := ReadSymbolTable;

  {open Con for console output}
  Assign(Con, '');
  SetTextBuf(Con, ConBuf, 1);
  Rewrite(Con);
  ConHandle := StdErrHandle;

  WriteLn(Con, SignOn);
  Initialize;

  if DoSymbolNames or DoProcSizes or DoVarTypes
  or DoInterface or DoImplement or not(DoDump) then
    DumpTPU;
  if DoDump then
    DumpRawData;
end.
