{$S-,R-,I-,V-,B-,F-}
{$I TADEFINE.INC}
{*********************************************************}
{*                    UNASM.PAS 7.00                     *}
{*   Copyright (c) 1988,1993 by TurboPower Software.     *}
{*                 All rights reserved.                  *}
{*    Portions copyright (c) 1987 by L. David Baldwin    *}
{*********************************************************}

unit Unasm;
  {-Unassemble code}

interface

uses
  {$IFDEF UseTPro}
  TpString;
  {$ELSE}
  OpString;
  {$ENDIF}

  procedure Unassemble(Start, Count : Word; InterpretNullCalls : Boolean;
                       Buffer : Pointer; var OutFile : Text);
    {-Unassemble code in Buffer, sending output to OutFile. Start is the
     starting IP, and Count is the number of bytes to unassemble}

implementation

{$DEFINE DoingTPeek}

const
  ULength = 127;
  CodeTab = 11;
  FirstTab = 24;
  SecondTab = 32;
type
  BigArray = array[0..32760] of Word;
  BigPtr = ^BigArray;
  Line = string[ULength];    {Disassembled instruction is built up in a 'line'}
  String127 = string[127];
  RegStrType = array[0..15] of array[1..2] of Char;
  SegRegType = array[0..3] of array[1..2] of Char;

  {Packet holds a displacement. It may be of byte or word size}
  Packet = record
             DispSize : (ByteSize, WordSize);
             Value : Word;
           end;
var
  InPtr : BigPtr;
  UString : Line;
  ULen : Byte absolute UString;
  UsIndex,
  CodeIndex,
  StopPC,
  OrigPC,
  PC : Word;
  InterpretNulls,
  AllDone,
  Wd,
  PrefixFl,
  Wait_Found : Boolean;
  Reg,
  Mode,
  Rm : Word;
  NValue,
  OpCode : Byte;
  OutF : ^Text;

const
  HexDigits : array[0..15] of Char = '0123456789ABCDEF';


  InstrNames : array[0..112] of string[6] = ( {6.00; various changes for 286}
    {00..07}  'AAA',  'AAD',  'AAM',  'AAS',  'ADC',  'ADD',  'AND',  'CALL',
    {08..15}  'CBW',  'CLC',  'CLD',  'CLI',  'CMC',  'CMP',  'CMPSB','CMPSW',
    {16..23}  'CS:',  'CWD',  'DAA',  'DAS',  'DB',   'DEC',  'DIV',  'DS:',
    {24..31}  'ES:',  'HLT',  'IDIV', 'IMUL', 'IN',   'INC',  'INT',  'INTO',
    {32..39}  'INT 3','IRET', 'JB',   'JBE',  'JCXZ', 'JZ',   'JL',   'JLE',
    {40..47}  'JMP',  'JNB',  'JA',   'JNZ',  'JGE',  'JG',   'JNO',  'JPO',
    {48..55}  'JNS',  'JO',   'JPE',  'JS',   'LAHF', 'LDS',  'LEA',  'LES',
    {56..63}  'LOCK', 'LODSB','LODSW','LOOP', 'LOOPE','LOOPNE','MOV', 'MOVSB',
    {64..71}  'MOVSW','MUL',  'NEG',  'NOP',  'NOT',  'OR',   'OUT',  'POP',
    {72..79}  'POPF', 'PUSH', 'PUSHF','???',  'RCL',  'RCR',  'REPE', 'REPNE',
    {80..87}  'RET',  'RETF', 'ROL',  'ROR',  'SAHF', 'SAR',  'SBB',  'SCASB',
    {88..95}  'SCASW','SHL',  'SHR',  'SS:',  'STC',  'STD',  'STI',  'STOSB',
    {96..103} 'STOSW','SUB',  'TEST', 'FWAIT','XCHG', 'XLAT', 'XOR',  'PUSHA',
    {104.111} 'POPA', 'BOUND','INSB', 'INSW', 'OUTSB','OUTSW','ENTER','LEAVE',
    {112    } 'ARPL');

  OpCodes : array[0..$FF] of Byte = ( {Index into InstrNames table}
    {00..0F}  05, 05, 05, 05, 05, 05, 73, 71, 69, 69, 69, 69, 69, 69, 73, 32,
    {10..1F}  04, 04, 04, 04, 04, 04, 73, 71, 86, 86, 86, 86, 86, 86, 73, 71,
    {20..2F}  06, 06, 06, 06, 06, 06, 24, 18, 97, 97, 97, 97, 97, 97, 16, 19,
    {30..3F} 102,102,102,102,102,102, 91, 00, 13, 13, 13, 13, 13, 13, 23, 03,
    {40..4F}  29, 29, 29, 29, 29, 29, 29, 29, 21, 21, 21, 21, 21, 21, 21, 21,
    {50..5F}  73, 73, 73, 73, 73, 73, 73, 73, 71, 71, 71, 71, 71, 71, 71, 71,
    {60..6F} 103,104,105,112, 20, 20, 20, 20, 73, 27, 73, 27,106,107,108,109,
    {70..7F}  49, 46, 34, 41, 37, 43, 35, 42, 51, 48, 50, 47, 38, 44, 39, 45,
    {80..8F}  20, 20, 20, 20, 98, 98,100,100, 62, 62, 62, 62, 62, 54, 62, 71,
    {90..9F}  67,100,100,100,100,100,100,100, 08, 17, 07, 99, 74, 72, 84, 52,
    {A0..AF}  62, 62, 62, 62, 63, 64, 14, 15, 98, 98, 95, 96, 57, 58, 87, 88,
    {B0..BF}  62, 62, 62, 62, 62, 62, 62, 62, 62, 62, 62, 62, 62, 62, 62, 62,
    {C0..CF}  20, 20, 80, 80, 55, 53, 62, 62,110,111, 81, 81, 32, 30, 31, 33,
    {D0..DF}  20, 20, 20, 20, 02, 01, 20,101, 20, 20, 20, 20, 20, 20, 20, 20,
    {E0..EF}  61, 60, 59, 36, 28, 28, 70, 70, 07, 40, 40, 40, 28, 28, 70, 70,
    {F0..FF}  56, 20, 79, 78, 25, 12, 20, 20, 09, 92, 11, 94, 10, 93, 20, 20);

  BranchClass : array[0..$FF] of Byte = ( {Unassemble branches based on OpCode}
    {00..0F} 05, 05, 05, 05, 11, 11, 10, 10, 05, 05, 05, 05, 11, 11, 10, 39,
    {10..1F} 05, 05, 05, 05, 11, 11, 10, 10, 05, 05, 05, 05, 11, 11, 10, 10,
    {20..2F} 05, 05, 05, 05, 11, 11, 01, 00, 05, 05, 05, 05, 11, 11, 01, 00,
    {30..3F} 05, 05, 05, 05, 11, 11, 01, 00, 05, 05, 05, 05, 11, 11, 01, 00,
    {40..4F} 03, 03, 03, 03, 03, 03, 03, 03, 03, 03, 03, 03, 03, 03, 03, 03,
    {50..5F} 03, 03, 03, 03, 03, 03, 03, 03, 03, 03, 03, 03, 03, 03, 03, 03,
    {60..6F} 00, 00, 13, 40, 32, 32, 32, 32, 38, 33, 34, 33, 00, 00, 00, 00,
    {70..7F} 07, 07, 07, 07, 07, 07, 07, 07, 07, 07, 07, 07, 07, 07, 07, 07,
    {80..8F} 35, 35, 35, 35, 05, 05, 05, 05, 05, 05, 05, 05, 18, 13, 18, 22,
    {90..9F} 00, 04, 04, 04, 04, 04, 04, 04, 00, 00, 09, 25, 00, 00, 00, 00,
    {A0..AF} 12, 12, 12, 12, 00, 00, 00, 00, 11, 11, 00, 00, 00, 00, 00, 00,
    {B0..BF} 06, 06, 06, 06, 06, 06, 06, 06, 06, 06, 06, 06, 06, 06, 06, 06,
    {C0..CF} 36, 36, 23, 00, 13, 13, 21, 21, 37, 00, 23, 00, 00, 14, 00, 00,
    {D0..DF} 20, 20, 20, 20, 24, 24, 32, 00, 27, 28, 26, 29, 27, 30, 26, 31,
    {E0..EF} 07, 07, 07, 07, 16, 16, 17, 17, 08, 08, 09, 07, 16, 16, 15, 15,
    {F0..FF} 02, 32, 02, 02, 00, 00, 19, 19, 00, 00, 00, 00, 00, 00, 19, 19);

  Grp1_2Names : array[0..15] of Byte =
  (98, 75, 68, 66, 65, 27, 22, 26, 29, 21, 7, 7, 40, 40, 73, 75);

  ShiftNames : array[0..7] of Byte = (82, 83, 76, 77, 89, 90, 75, 85);

  ImmedNames : array[0..7] of Byte = (5, 69, 4, 86, 6, 97, 102, 13);

  RegStr : RegStrType = (
    'AX', 'CX', 'DX', 'BX', 'SP', 'BP', 'SI', 'DI',
    'AL', 'CL', 'DL', 'BL', 'AH', 'CH', 'DH', 'BH');
  SegRegStr : SegRegType = ('ES', 'CS', 'SS', 'DS');
  WordPtr : array[Boolean] of string[9] = ('BYTE PTR ', 'WORD PTR ');
  CsPending : Boolean = False;

  procedure InsertChr(C : Char);
  begin
    UString[UsIndex] := C;
    if ULen < UsIndex then
      ULen := UsIndex;
    Inc(UsIndex);
  end;

  procedure Comma;
  begin
    InsertChr(',');
  end;

  procedure InsertSt(S : String127);
  var
    I : Byte;
    SLen : Byte absolute S;
  begin
    for I := 1 to SLen do
      InsertChr(S[I]);
  end;

  procedure InsertHx2(B : Byte);
  begin
    InsertSt(HexB(B));
  end;

  procedure InsertOffset(O : Word);
  var
    St : String[9];
  begin
    {$IFDEF DoingTPeek}
      St := HexW(O-OrigPC)+'|'+HexW(O);
    {$ELSE}
      St := HexW(O);
    {$ENDIF}
    Move(St[1], UString[1], Length(St));
  end;

  procedure InsertCode(B : Byte);
  begin
    UString[CodeIndex] := HexDigits[B shr 4];
    UString[CodeIndex+1] := HexDigits[B and $F];
    Inc(CodeIndex, 2);
  end;

  procedure InsertHx4(W : Word);
  begin
    InsertSt(HexW(W));
  end;

  procedure InitLineBuffer;
  {-Initialize UString for next line}
  begin
    ULen := 0;
    FillChar(UString[1], ULength, ' '); {clear ustring}
    InsertOffset(PC);
    CodeIndex := CodeTab;
  end;

  procedure OutUString;
  begin
    WriteLn(OutF^, UString);
    InitLineBuffer;
  end;

  procedure Next;
  type
    A2 = array[-1..0] of Byte;
  var
    P : ^A2;
    POfs : Word absolute P;
  begin
    P := Pointer(InPtr);
    Inc(POfs, PC-Succ(OrigPC));
    NValue := P^[0];
    InsertCode(P^[-1]);
    if PC > StopPC then
      AllDone := True;
  end;

  procedure GetByte(var P : Packet);
  begin
    P.DispSize := ByteSize;
    P.Value := NValue;
    Inc(PC);
    Next;
  end;

  procedure GetWord(var P : Packet);
  var
    H, L : Packet;
  begin
    P.DispSize := WordSize;
    GetByte(L);
    GetByte(H);
    P.Value := (H.Value shl 8)+L.Value;
  end;

  procedure InsertDisp(Disp : Packet; Sign : Boolean);
  begin
    with Disp do
      if (DispSize = ByteSize) then begin
        if Sign then
          if Value and $80 <> 0 then begin
            InsertChr('-');  {turn into negative number}
            Value := -(Value or $FF00);
          end
          else
            InsertChr('+');
        InsertHx2(Lo(Value));
      end
      else
        InsertHx4(Value);
  end;

  procedure InsertLabel(N : Word);
  begin
    InsertHx4(N);
  end;

  procedure ShortJump;
    {the short jump instructions}
  var
    Pk : Packet;
    Vl : Word;
  begin
    GetByte(Pk);
    if (OpCode = $EB) then
      InsertSt('SHORT ');
    with Pk do begin
      Vl := Value;
      if (Vl and $80 <> 0) then
        Vl := Vl or $FF00;   {sign extend}
      InsertLabel(Vl+PC);
    end;
  end;

  procedure IntraSeg;
    {the intrasegment direct jumps and calls}
  var
    Pk : Packet;
  begin
    GetWord(Pk);

    if InterpretNulls then
      Inc(Pk.Value, PC)
    else
      if Pk.Value <> 0 then
        Inc(Pk.Value, PC);

    InsertLabel(Pk.Value);
  end;

  procedure InterSeg;
    {-The intersegment direct jumps and calls}
  var
    S, O : Packet;
  begin                      {InterSeg}
    GetWord(O);
    GetWord(S);
    InsertSt('FAR '+HexW(S.Value)+':'+HexW(O.Value));
  end;                       {InterSeg}

  procedure MovImToReg;
    {the move immediate to a reg such as mov bl,12 }
  var
    Disp : Packet;
  begin
    Reg := (OpCode and $F) xor 8;
    InsertSt(RegStr[Reg]);
    Comma;
    if (OpCode and 8) <> 0 {word} then
      GetWord(Disp)
    else
      GetByte(Disp);
    InsertDisp(Disp, False);
  end;

  procedure DoMem(Disp : Packet);
  const
    RegPhrase : array[0..7] of string[5] =
    ('BX+SI', 'BX+DI', 'BP+SI', 'BP+DI', 'SI', 'DI', 'BP', 'BX');
  begin
    if Mode = 3 then begin   {its a reg}
      if not Wd then
        Rm := Rm+8;
      InsertSt(RegStr[Rm]);
    end
    else begin               {its a memory}
      if CsPending then begin
        InsertSt('CS:');
        CsPending := False;
      end;
      InsertChr('[');
      if (Rm = 6) and (Mode = 0) then
        InsertDisp(Disp, False)
      else begin             {need a register phrase}
        InsertSt(RegPhrase[Rm]);
        if Mode <> 0 then begin
          if (Disp.DispSize = WordSize) then
            InsertChr('+');
          InsertDisp(Disp, True);
        end;
      end;
      InsertChr(']');
    end;
  end;

  procedure DoReg;
  begin
    if not Wd then
      Inc(Reg, 8);
    InsertSt(RegStr[Reg]);
  end;

  procedure ReadModeByte(var Disp : Packet);
    {-read the mode byte and sort out the various parts.  Read the
      displacement byte or word if req'D}
  var
    ModeByte : Byte;
    Pk : Packet;
  begin
    GetByte(Pk);
    ModeByte := Lo(Pk.Value);
    Rm := ModeByte and 7;
    Mode := (ModeByte and $C0) shr 6 {Div 64} ;
    Reg := (ModeByte and $38) shr 3 {Div 8} ;
    if (Mode = 0) and (Rm = 6) or (Mode = 2) then
      GetWord(Disp)          {get address or 16 bit disp}
    else if Mode = 1 then    {its a 8 bit displ}
      GetByte(Disp);
  end;

  procedure MemSeg;
    {-Move seg reg to/from mem/reg}
  var
    Disp : Packet;
    ToReg : Boolean;
  begin                      {MemSeg}
    ToReg := (OpCode and 2) <> 0;
    Wd := True;
    ReadModeByte(Disp);
    Reg := Reg and 3;        {0..3}
    if ToReg then
      InsertSt(SegRegStr[Reg])
    else
      DoMem(Disp);
    Comma;
    if ToReg then
      DoMem(Disp)
    else
      InsertSt(SegRegStr[Reg]);
  end;                       {MemSeg}

  procedure ImedToAc;
    {-do the immediates to ac}
  var
    Disp : Packet;
  begin
    Wd := (OpCode and 1) <> 0;
    Reg := 0;                {ax or al}
    if Wd then
      GetWord(Disp)
    else
      GetByte(Disp);
    DoReg;
    Comma;
    if Wd then
      InsertDisp(Disp, False)
    else
      InsertHx2(Lo(Disp.Value)); {no sign}
  end;

  procedure Immed;
    {-add reg/mem,12   xor reg/mem,1234}
  var
    SignExt : Boolean;
    D1, D2 : Packet;
  begin
    Wd := (OpCode and 1) <> 0;
    SignExt := ((OpCode and 2) <> 0) and (OpCode <= $83); {mov does not have sign ext}
    ReadModeByte(D1);
    if OpCode <= $83 then    {mov has name output already}
      InsertSt(InstrNames[ImmedNames[Reg]]);
    UsIndex := SecondTab;
    if Wd and not SignExt then
      GetWord(D2)
    else
      GetByte(D2);
    if Mode <> 3 then
      InsertSt(WordPtr[Wd]);
    DoMem(D1);
    Comma;
    InsertDisp(D2, False);
  end;

  procedure DoEsc(Pk : Packet);
    {-do the esc instr}
  begin
    if Wait_Found then begin
      InsertSt(InstrNames[OpCodes[$9B]]);
      OutUString;
    end;
    UsIndex := FirstTab;
    InsertSt('ESC');
    UsIndex := SecondTab;
    InsertHx2((OpCode and 7)*8+Reg);
    Comma;
    DoMem(Pk);
  end;

  procedure St_I;
    {-do st(i) }
  begin
    InsertSt('ST('+Chr(Rm+$30)+')');
  end;

  procedure Sti_St;
    {-do st(i),st }
  begin
    St_I;
    InsertSt(',ST');
  end;

  procedure Db;
  const
    Db_Name : array[0..11] of string[6] = (
     'FILD', '', 'FIST', 'FISTP', '', 'FLD', '', 'FSTP', 'FENI', 'FDISI',
     'FCLEX', 'FINIT');
  var
    I : Word;
    Pk : Packet;
    Tmp : string[7];
  begin
    ReadModeByte(Pk);
    Wd := False;
    if (Mode = 3) then
      I := Rm+8
    else
      I := Reg;              {form an index}
    Tmp := Db_Name[I];
    if (Tmp[0] > #0) and (I <= 11) then begin
      if not Wait_Found then
        if I >= 8 then       {'FENI','FDISI','FCLEX','FINIT' to no wait equivalents}
          Insert('N', Tmp, 2);
      InsertSt(Tmp);
      UsIndex := SecondTab;
      if I <= 3 then begin
        InsertSt('DWORD ');
        DoMem(Pk);
      end
      else
        if I <= 7 then begin
          InsertSt('TBYTE ');
          DoMem(Pk);
        end;
    end
    else
      DoEsc(Pk);
  end;

  procedure Dd;
  const
    Dd_Name : array[0..11] of string[6] =
    ('FLD', '', 'FST', 'FSTP', 'FRSTOR', '', 'FSAVE', 'FSTSW', 'FFREE',
      'FXCH', 'FST', 'FSTP');
  var
    I : Word;
    Pk : Packet;
    Tmp : string[7];
  begin
    ReadModeByte(Pk);
    Wd := False;
    if Mode = 3 then
      I := Reg+8
    else
      I := Reg;
    Tmp := Dd_Name[I];
    if (Tmp[0] > #0) and (I <= 11) then begin
      if not Wait_Found then
        if (I = 6) or (I = 7) then {'FSAVE','FSTSW' to no wait equivalents}
          Insert('N', Tmp, 2);
      InsertSt(Tmp);
      UsIndex := SecondTab;
      if I <= 3 then
        InsertSt('QWORD ');
      if I <= 7 then
        DoMem(Pk)
      else
        St_I;
    end
    else
      DoEsc(Pk);
  end;

  procedure DF;
    {-Do $DF opcode}
  const
    Df_Name : array[0..11] of string[5] =
    ('FILD', '', 'FIST', 'FISTP', 'FBLD', 'FILD', 'FBSTP', 'FISTP', 'FFREE',
      'FXCH', 'FST', 'FSTP');
    TQ : array[Boolean] of string[6] =
    ('QWORD ', 'TBYTE ');
  var
    I : Word;
    Pk : Packet;
  begin                      {DF}
    ReadModeByte(Pk);
    Wd := False;
    if Mode = 3 then
      I := Reg+8
    else
      I := Reg;              {form index}
    if (I <> 1) and (I <= 11) then begin
      InsertSt(Df_Name[I]);
      UsIndex := SecondTab;
      case I of
        2..3 : begin
                 InsertSt('WORD ');
                 DoMem(Pk);
               end;
        4..7 : begin
                 InsertSt(TQ[(I and 5) = 4]);
                 DoMem(Pk);
               end
        else St_I;
      end;
    end
    else
      DoEsc(Pk);
  end;                       {DF}

  procedure D9;
  const
    D9_Name : array[0..11] of string[6] =
    ('FLD', '', 'FST', 'FSTP', 'FLDENV', 'FLDCW', 'FSTENV', 'FSTCW',
      'FLD', 'FXCH', 'FNOP', 'FSTP');

    D9_3_Name : array[0..29] of string[7] =
    ('FCHS', 'FABS', '', '', 'FTST', 'FXAM', '', '', 'FLD1', 'FLDL2T',
      'FLDL2E', 'FLDPI', 'FLDLG2', 'FLDLN2', 'FLDZ', '', 'F2XM1',
      'FYL2X', 'FPTAN', 'FPATAN', 'FXTRACT', '', 'FDECSTP', 'FINCSTP',
      'FPREM', 'FYL2XP1', 'FSQRT', '', 'FRNDINT', 'FSCALE');
  var
    I : Word;
    Pk : Packet;
    Tmp : string[7];
  begin
    ReadModeByte(Pk);
    Wd := False;
    if (Mode <> 3) or (Reg <= 3) then begin
      if Mode = 3 then
        I := Reg+8
      else
        I := Reg;
      if (I = 1) or ((I = 10) and (Rm <> 0)) then
        DoEsc(Pk)
      else begin
        Tmp := D9_Name[I];
        if not Wait_Found then
          if (I = 6) or (I = 7) then {'FSTENV', 'FSTCW' to no wait equivalents}
            Insert('N', Tmp, 2);
        InsertSt(Tmp);
        UsIndex := SecondTab;
        if I <= 3 then
          InsertSt('DWORD ');
        if I <= 7 then
          DoMem(Pk)
        else
          if I <> 10 then    {fnop is 10}
            St_I;            {st(i)}
      end;
    end
    else begin               {mode=3 and reg>=4}
      I := Rm+(Reg and 3)*8; {include lower 2 bits of reg in index}
      if (D9_3_Name[I][0] > #0) and (I <= 29) then
        InsertSt(D9_3_Name[I])
      else
        DoEsc(Pk);
    end;
  end;

  procedure D8_Dc;
  type
    NameType = array[0..7] of array[1..5] of Char;
  var
    ShortReal : Boolean;
    Pk : Packet;
  const
    D8_Dc_Name : NameType =
    ('FADD ', 'FMUL ', 'FCOM ', 'FCOMP', 'FSUB ', 'FSUBR', 'FDIV ', 'FDIVR');
  begin
    Wd := False;
    ShortReal := OpCode = $D8;
    ReadModeByte(Pk);
    if not ShortReal then
      if (Reg >= 6) then     {fdiv, fdivr are reversed here}
        Reg := Reg xor 1;
    InsertSt(D8_Dc_Name[Reg]);
    UsIndex := SecondTab;
    if Mode <> 3 then begin
      if ShortReal then
        InsertSt('DWORD ')
      else
        InsertSt('QWORD ');
      DoMem(Pk);
    end
    else                     {mode=3}
      if ShortReal then begin
        InsertSt('ST,');
        St_I;
      end
      else
        Sti_St;              {add the stack info}
  end;

  procedure Da_De;
  type
    NameType = array[0..15] of array[1..6] of Char;
  var
    ShortInt : Boolean;
    Pk : Packet;
  const
    Da_De_Name : NameType = (
      'FIADD ', 'FIMUL ', 'FICOM ', 'FICOMP', 'FISUB ', 'FISUBR', 'FIDIV ',
      'FIDIVR', 'FADDP ', 'FMULP ', 'FCOMP ', 'FCOMPP', 'FSUBRP', 'FSUBP ',
      'FDIVRP', 'FDIVP ');
  begin
    Wd := False;
    ShortInt := OpCode = $DA;
    ReadModeByte(Pk);
    if Mode <> 3 then begin
      InsertSt(Da_De_Name[Reg]);
      UsIndex := SecondTab;
      if ShortInt then
        InsertSt('DWORD ')
      else
        InsertSt('WORD ');
      DoMem(Pk);
    end
    else begin               {mode=3}
      if ShortInt or ((Reg = 3) and (Rm <> 1)) then
        DoEsc(Pk)            {not fl pt}
      else begin
        InsertSt(Da_De_Name[Reg+8]);
        UsIndex := SecondTab;
        if (Reg <> 3) then
          Sti_St;
      end;
    end;
  end;

  procedure DoShift;
    {-do the shift and rotate instr}
  var
    Pk : Packet;
  begin
    Wd := (OpCode and 1) <> 0;
    ReadModeByte(Pk);
    InsertSt(InstrNames[ShiftNames[Reg]]);
    UsIndex := SecondTab;
    if Mode <> 3 then
      InsertSt(WordPtr[Wd]);
    DoMem(Pk);
    Comma;
    if (OpCode and 2) <> 0 then
      InsertSt('CL')
    else
      InsertSt('1');
  end;

  procedure DoGroup1_2;
    {-f6,f7,fe,ff}
  var
    Pk : Packet;
  begin
    Wd := (OpCode and 1) <> 0;
    ReadModeByte(Pk);
    if (OpCode and 8) <> 0 then
      Reg := Reg+8;
    {reg is ptr to name in this case}
    if (OpCode = $FE) then
      if (Reg >= $A) then
        Reg := $F;

    {no call, jmp, push of bytes}
    InsertSt(InstrNames[Grp1_2Names[Reg]]);
    UsIndex := SecondTab;
    case Reg of
      $A, $C : {InsertSt('NEAR ')} ;
      $B, $D : InsertSt('FAR ');
    else
      if (Mode <> 3) then
        if (Reg <> $E) {push} then
          InsertSt(WordPtr[Wd]);
    end;
    DoMem(Pk);
    if Reg = 0 then begin    {test}
      Comma;
      if Wd then begin
        GetWord(Pk);
        InsertDisp(Pk, False);
      end
      else begin
        GetByte(Pk);
        InsertHx2(Lo(Pk.Value)); {no sign}
      end;
    end;
  end;

  procedure MemToReg;
    {-lds,les,lea}
  var
    Pk : Packet;
  begin
    Wd := True;
    ReadModeByte(Pk);
    DoReg;
    Comma;
    DoMem(Pk);
  end;

  procedure MemAccum;
    {-handle mov ac,[1234] , cmp ac,[5678] etc}
  var
    Disp : Packet;
    ToReg : Boolean;
  begin
    Wd := (OpCode and 1) <> 0;
    ToReg := (OpCode and 2) = 0; {note the difference in sense}
    Reg := 0;                {will be AX or AL}
    GetWord(Disp);
    Rm := 6;
    Mode := 0;               {for displacement only}
    if ToReg then
      DoReg
    else
      DoMem(Disp);
    Comma;
    if ToReg then
      DoMem(Disp)
    else
      DoReg;
  end;

  procedure MRegMReg;
    {-do the mem/reg, mem/reg instructions, such as mov bx,[bp+1234]
    or add [bx],dx }
  var
    Pk : Packet;
    ToReg : Boolean;
  begin
    Wd := (OpCode and 1) <> 0;
    ToReg := (OpCode and 2) <> 0;
    ReadModeByte(Pk);
    if ToReg then
      DoReg
    else
      DoMem(Pk);
    Comma;
    if ToReg then
      DoMem(Pk)
    else
      DoReg;
  end;

  procedure RegMRegImed;
  {- do RegW,memW/RegW,Immed. types}
  var
    Pk : Packet;
  begin
    Wd := True;
    ReadModeByte(Pk);
    DoReg;
    Comma;
    DoMem(Pk);
    Comma;
    GetWord(Pk);
    InsertDisp(Pk, False);
  end;

  procedure Rep_Lock;
    {-do lock, repe, repne,wait, and seg overrides}
  begin
    PrefixFl := True;
    OutUString;
  end;

  procedure InsByte;
  var
    Pk : Packet;
  begin
    GetByte(Pk);
    InsertHx2(Lo(Pk.Value));
  end;

  function DoInt(NVal : Byte) : Boolean;
  var
    Pk : Packet;
    OC : Byte;
  begin
    {Emulated 87 instruction?}
    DoInt := False;
    case NVal of
      $34..$3D :
        begin
          GetByte(Pk);
          case NVal OF
            $34 : OpCode := $D8;
            $35 : OpCode := $D9;
            $36 : OpCode := $DA;
            $37 : OpCode := $DB;
            $38 : OpCode := $DC;
            $39 : OpCode := $DD;
            $3A : OpCode := $DE;
            $3B : OpCode := $DF;
            $3C : begin
                    CsPending := True;
                    OC := ((NValue and $0F) or $D0);
                    GetByte(Pk);
                    OpCode := OC;
                  end;
            $3D : begin
                    InsertSt(InstrNames[OpCodes[$9B]]);
                    DoInt := True;
                  end;
          end;
        end;
      else begin
        {plain old INT nn}
        InsertSt(InstrNames[OpCodes[$CD]]);
        UsIndex := SecondTab;
        InsByte;
        DoInt := True;
      end;
    end;
  end;

  procedure ImmedByte;
    {-do immediate byte}
  var
    D : Packet;
  begin
    UsIndex := SecondTab;
    GetByte(D);
    InsertDisp(D, False);
  end;

  procedure ImmedWord;
    {-do immediate word}
  var
    D : Packet;
  begin
    UsIndex := SecondTab;
    GetWord(D);
    InsertDisp(D, False);
  end;

  procedure ImmedWordImmedByte;
    {-do immediate word,immediate byte}
  var
    D : Packet;
  begin
    ImmedWord;
    Comma;
    GetByte(D);
    InsertDisp(D, False);
  end;

  procedure RMbwRMbwIbw;
  {-do Reg/Mem-byte/word,Reg/Mem-byte/word or Immediate-byte/word}
  const
    InstrIndx : array[0..7] of byte =
     (5,69,4,86,6,97,102,13);
  var
    Pk : Packet;
  begin
    ReadModeByte(Pk);
    InsertSt(InstrNames[InstrIndx[Reg]]);
    UsIndex := SecondTab;
    Wd := (OpCode and 1) <> 0;
    if Mode <> 3 then
      InsertSt(WordPtr[Wd]);
    DoMem(Pk);
    Comma;
    if OpCode=$81 then
      GetWord(Pk)
    else
      GetByte(Pk);
    InsertDisp(Pk, False);
  end;

  procedure RMbwIb;
  {-do Reg/Mem-byte/word,Immediate byte}
  const
    InstrIndx : array[0..7] of byte =
     (82,83,76,77,89,90,89,85);
  var
    Pk : Packet;
  begin
    ReadModeByte(Pk);
    InsertSt(InstrNames[InstrIndx[Reg]]);
    UsIndex := SecondTab;
    Wd := (OpCode and 1) <> 0;
    if Mode <> 3 then
      InsertSt(WordPtr[Wd]);
    DoMem(Pk);
    Comma;
    GetByte(Pk);
    InsertDisp(Pk, False);
  end;

  procedure DoDBFromBuffer(N:Byte);
  {-Report N bytes as DB from line buffer}
  var
    i : Byte;
  begin
    InsertSt(InstrNames[20]);
    UsIndex := SecondTab;
    for i := 1 to pred(N) do begin
      InsertSt(Copy(UString,11+pred(i)*2,2));
      Comma;
    end;
    InsertSt(Copy(UString,11+pred(N)*2,2));
  end;

  procedure Do0FhCodes;
    {-do codes starting with $0F}
  const
    LarLsl : array[2..3] of String[3] = ('LAR','LSL');
    Clts   : String[4] = 'CLTS';
    T_0F_00: array[0..5] of String[4] =
             ('SLDT','STR','LLDT','LTR','VERR','VERW');
    T_0F_01: array[0..6] of String[4] =
             ('SGDT','SIDT','LGDT','LIDT','SMSW','','LMSW');
  var
    Pk : Packet;
  begin
    GetByte(Pk);
    case Pk.Value of
    0 :
      begin
        ReadModeByte(Pk);
        if Reg > 5 then {Not supported, report as DB and get out}
          DoDBFromBuffer(3)
        else
          begin
            InsertSt(T_0F_00[Reg]);
            UsIndex := SecondTab;
            Wd := True;
            DoMem(Pk);
          end;
      end;
    1 :
      begin
        ReadModeByte(Pk);
        if (Reg = 5) or (Reg = 7) then
          DoDBFromBuffer(3) {Not supported, report as DB and get out}
        else
          begin
            InsertSt(T_0F_01[Reg]);
            UsIndex := SecondTab;
            Wd := True;
            DoMem(Pk);
          end;
      end;
    2,3 :
      begin
        InsertSt(LarLsl[Pk.Value]);
        UsIndex := SecondTab;
        MemToReg;
      end;
    6 : InsertSt(Clts);

    else {Instruction type not supported, report as DBs}

      DoDBFromBuffer(2);

    end;
  end;

  procedure MemReg;
  {-do Mem,Mem/Reg types}
  var
    Pk : Packet;
  begin
    ReadModeByte(Pk);
    DoMem(Pk);
    Comma;
    DoReg;
  end;

  procedure UnassemblePrim;
    {-Unassemble one line of code (or two if preceeded by a seg instruction)
      output the unassembled line in UString.}
  label
    98, 99;
  var
    Pk : Packet;
  begin                      {UnassemblePrim}
    Wait_Found := False;
    InitLineBuffer;
    repeat
      PrefixFl := False;     {set true later if a segm overide instr found}
      GetByte(Pk);
      OpCode := Pk.Value;
      UsIndex := FirstTab;
      case OpCode of
        $0..$E,
        $10..$7F,
        $99..$BF, $C2..$CC,
        $CE..$CF : {skip other checks} ;
        $0F,
        $80..$83, $C0, $C1,
        $9B, $CD,
        $D0..$D3, $D8..$DF,
        $F6..$F7, $FE..$FF : goto 98; { Do later }
      end;

      {most items have opcode name output now}
      InsertSt(InstrNames[OpCodes[OpCode]]);
      UsIndex := SecondTab;
98:
      case BranchClass[OpCode] of
        0 : {opcode only} ;
        1,                   {seg overide instruction}
        2 : Rep_Lock;        {LOCK, REPE, REPNE}
        3 : InsertSt(RegStr[OpCode and 7]); {PUSH, POP, XCHG, INC, DEC}
        4 : InsertSt(RegStr[OpCode and 7]+',AX'); {XCHG ??,AX}
        5 : MRegMReg;
        6 : MovImToReg;      {MOV CX,1234 etc.}
        7 : ShortJump;
        8 : IntraSeg;
        9 : InterSeg;
        10 :                 {SEG, PUSH-POP SEG}
          InsertSt(SegRegStr[(OpCode shr 3) and 3]);
        11 : ImedToAc;
        12 : MemAccum;       {MOV AC,[1234] }
        13 : MemToReg;       {LES,LDS,LEA}
        14 : if not DoInt(NValue) then {INT N}
               goto 98;
        15 : begin           {OUT DX,AC}
               Wd := True;
               Reg := 2;
               DoReg;
99:            Comma;
               Wd := (OpCode and 1) <> 0;
               Reg := 0;     {AX OR AL}
               DoReg;
             end;
        16 : begin           {IN AC, DX or PORT}
               Wd := (OpCode and 1) <> 0;
               Reg := 0;
               DoReg;
               Comma;
               if (OpCode >= $EC) then begin
                 Wd := True;
                 Reg := 2;
                 DoReg;
               end
               else
                InsByte;
             end;
        17 : begin           {OUT PORT,AC}
               InsByte;
               goto 99;
             end;
        18 : MemSeg;         {SEGMENT, REG INSTR}
        19 : DoGroup1_2;
        20 : DoShift;
        21 : Immed;
        22 : begin           {POP REG/MEM}
               Wd := True;
               ReadModeByte(Pk);
               DoMem(Pk);
             end;
        23 : begin           {RET N}
               GetWord(Pk);
               InsertDisp(Pk, False);
             end;
        24 : begin           {AAM,AAD}
               GetByte(Pk);
               if (Pk.Value <> $A) then
                 InsertHx2(Lo(Pk.Value))
             end;
        25 : if (NValue >= $D8) and (NValue <= $DF) then begin
               {WAIT - look to see if it preceeds a Fl Point instr}
               Wait_Found := True;
               PrefixFl := True;
             end
             else            {plain WAIT}
               InsertSt(InstrNames[OpCodes[$9B]]);
        26 : Da_De;
        27 : D8_Dc;
        28 : D9;
        29 : Db;
        30 : Dd;
        31 : DF;
        32 : InsertHx2(OpCode); {for DB (data byte)}
        33 : RegMRegImed;
        34 : ImmedByte;
        35 : RMbwRmbwIbw;
        36 : RMbwIb;
        37 : ImmedWordImmedByte;
        38 : ImmedWord;
        39 : Do0FhCodes;
        40 : MemReg;
      end;
    until (PrefixFl = False);
    OutUString;
  end;                       {UnassemblePrim}

  procedure Unassemble(Start, Count : Word; InterpretNullCalls : Boolean;
                       Buffer : Pointer; var OutFile : Text);
    {-Unassemble code in Buffer, sending output to OutFile. Start is the
     starting IP, and Count is the number of bytes to unassemble}
  begin                      {Unassemble}
    OutF := @OutFile;
    InPtr := Buffer;
    InterpretNulls := InterpretNullCalls;
    PC := Start;
    OrigPC := PC;
    StopPC := Pred(PC+Count);
    AllDone := False;
    CodeIndex := CodeTab;
    Next;
    while not AllDone do
      UnassemblePrim;
  end;                       {Unassemble}

end.
